import os


class FileMan(object):
    """Define a File Manager class for PowerSystem"""

    def __init__(self,
                 case,
                 input_format=None,
                 addfile=None,
                 config=None,
                 no_output=False,
                 dynfile=None,
                 dump_raw=None,
                 output_format=None,
                 output=None,
                 pert=None,
                 **kwargs):
        """initialize the output file names

        case: must be full path to case
        fullname: full name of case only
        name: name of case WITHOUT extension

        dump: desired simulation result file name
        output: desired name for format conversion output
        """
        if isinstance(input_format, str):
            self.input_format = input_format.lower()
        else:
            self.input_format = None
        if isinstance(output_format, str):
            self.output_format = output_format.lower()
        else:
            self.output_format = None

        self.case = case
        path, fullname = os.path.split(case)
        self.fullname = fullname
        self.name, self.ext = os.path.splitext(fullname)
        if not path:
            self.path = os.getcwd()
        else:
            self.path = path

        self.addfile = self.get_fullpath(addfile)
        self.pert = pert
        self.dynfile = self.get_fullpath(dynfile)
        self.config = self.get_fullpath(config)
        self.add_format = None

        if no_output:
            self.no_output = True
            self.log = None
            self.output = None
            self.lst = None
            self.eig = None
            self.dat = None
            self.dump_raw = None
            self.prof = None
        else:
            self.no_output = False
            if not output:
                output = add_suffix(self.name, 'out')
            if not dump_raw:
                dump_raw = add_suffix(self.name, 'dm')
            prof = add_suffix(self.name, 'prof')
            eig = add_suffix(self.name, 'eig')

            self.lst = add_ext(output, 'lst')
            self.eig = add_ext(eig, 'txt')
            self.dat = add_ext(output, 'dat')
            self.output = add_ext(output, 'txt')
            self.dump_raw = add_ext(dump_raw, 'dm')
            self.prof = add_ext(prof, 'txt')

    def get_fullpath(self, fullname=None):
        """
        Return the original full path if full path is specified, otherwise
        search in the case file path
        """
        if not fullname:
            return None

        path, name = os.path.split(fullname)
        if not name:
            return None
        else:
            if not path:
                return os.path.join(self.path, name)
            else:
                return os.path.join(path, name)


def add_ext(name, extension):
    """Add extension to a name, discard the previos extension if exists"""
    name, ext = os.path.splitext(name)
    return name + '.' + extension


def add_suffix(fullname, suffix):
    """ Add suffix to a full file name"""
    name, ext = os.path.splitext(fullname)
    return name + '_' + suffix + ext



from .base import ModelBase


class Recorder(ModelBase):
    """
    Recorder class that outputs simulation data of the given model, variable
    and elements
    """

    def define(self):
        self._group = 'Output'
        self.param_remove('Sn')
        self.param_remove('Vn')
        self.param_remove('fn')

        self.param_define(
            'model',
            default='ALL',
            tomatrix=False,
            descr='model to record',
            mandatory=True)
        self.param_define(
            'variable',
            default='ALL',
            tomatrix=False,
            descr='variable to record',
            mandatory=False)
        self.param_define(
            'element',
            default='ALL',
            tomatrix=False,
            descr='element idx to record',
            mandatory=False)

        self.calls.update({'init1': True})

        self.varout_idx = []
        self.varout_state_idx = []
        self.varout_algeb_idx = []
        self._init()

    def elem_add(self, idx=None, name=None, **kwargs):
        super(Recorder, self).elem_add(idx, name, **kwargs)

    def init1(self, dae):
        if not self.n:
            return

        # include line flow variables in algebraic variables
        nflows = 0
        if self.system.tds.config.compute_flows:
            nflows = 2 * self.system.Bus.n + \
                     8 * self.system.Line.n + \
                     2 * self.system.Area.n_combination

        if 'ALL' in self.model:
            self.varout_idx = list(range(dae.m + dae.n + nflows))
            return

        for model, variable, element in zip(self.model, self.variable,
                                            self.element):
            if isinstance(model, str):
                model = [model]

            # for each model
            for m in model:
                assert m in self.system.devman.devices

                m_instance = self.system.__dict__[m]
                offset = [0]
                if variable == 'ALL':
                    variable = m_instance._states + m_instance._algebs
                    offset = [0] * len(m_instance._states) + [
                        self.system.dae.n
                    ] * len(m_instance._algebs)
                elif isinstance(variable, str):
                    if variable in m_instance._algebs:
                        offset = [self.system.dae.n]
                    variable = [variable]

                # for each variable
                for v, offs in zip(variable, offset):
                    if element == 'ALL':
                        idx = self.system.__dict__[m].get_field(v)
                    else:
                        idx = self.system.__dict__[m].get_field(v, idx=element)

                    idx = [i + offs for i in idx]

                    self.varout_idx.extend(idx)



import importlib
import logging

logger = logging.getLogger(__name__)


class JIT(object):
    """
    Dummy Just-in-Time initialization class
    """

    def __init__(self, system, model, device, name):
        """constructor of a dummy JIT class with basic information"""
        self.system = system
        self.model = model
        self.device = device
        self.name = name
        self.loaded = 0

    def jit_load(self):
        """
        Import and instantiate this JIT object

        Returns
        -------

        """
        try:
            model = importlib.import_module('.' + self.model, 'andes.models')
            device = getattr(model, self.device)
            self.system.__dict__[self.name] = device(self.system, self.name)

            g = self.system.__dict__[self.name]._group
            self.system.group_add(g)
            self.system.__dict__[g].register_model(self.name)

            # register device after loading
            self.system.devman.register_device(self.name)
            self.loaded = 1
            logger.debug('Imported model <{:s}.{:s}>.'.format(
                self.model, self.device))
        except ImportError:
            logger.error(
                'non-JIT model <{:s}.{:s}> import error'
                .format(self.model, self.device))
        except AttributeError:
            logger.error(
                'model <{:s}.{:s}> not exist. Check models/__init__.py'
                .format(self.model, self.device))

    def __getattr__(self, attr):
        if not self.loaded:
            self.jit_load()
        return self.system.__dict__[self.name].__getattribute__(attr)

    def elem_add(self, idx=None, name=None, **kwargs):
        """overloading elem_add function of a JIT class"""
        self.jit_load()
        if self.loaded:
            return self.system.__dict__[self.name].elem_add(
                idx, name, **kwargs)

    def doc(self, **kwargs):
        self.jit_load()
        if self.loaded:
            return self.system.__dict__[self.name].doc(**kwargs)



""" Simple MATPOWER format parser
"""
import re
from ..consts import deg2rad

import logging
logger = logging.getLogger(__name__)


def testlines(fid):
    return True  # hard coded


def read(file, system):
    """Read a MATPOWER data file into mpc and build andes device elements"""
    func = re.compile('function\\s')
    mva = re.compile('\\s*mpc.baseMVA\\s*=\\s*')
    bus = re.compile('\\s*mpc.bus\\s*=\\s*\\[')
    gen = re.compile('\\s*mpc.gen\\s*=\\s*\\[')
    branch = re.compile('\\s*mpc.branch\\s*=\\s*\\[')
    area = re.compile('\\s*mpc.areas\\s*=\\s*\\[')
    gencost = re.compile('\\s*mpc.gencost\\s*=\\s*\\[')
    bus_name = re.compile('\\s*mpc.bus_name\\s*=\\s*\\{')
    end = re.compile('\\s*\\];?')
    hasdigit = re.compile('.*\\d+\\s*]?;?')
    comment = re.compile('\\s*%.*')

    retval = True
    field = None
    info = True

    basemva = 100
    mpc = {
        'bus': [],
        'gen': [],
        'branch': [],
        'area': [],
        'gencost': [],
        'bus_name': [],
    }

    fid = open(file, 'r')

    for line in fid:
        line = line.strip().rstrip(';')
        if not line:
            continue
        elif func.search(line):  # skip function declaration
            continue
        elif comment.search(line):  # for comment lines
            if info:
                logger.info(line[1:72])
                info = False
            else:
                continue
        elif mva.search(line):
            basemva = float(line.split('=')[1])

        if not field:
            if bus.search(line):
                field = 'bus'
            elif gen.search(line):
                field = 'gen'
            elif branch.search(line):
                field = 'branch'
            elif area.search(line):
                field = 'area'
            elif gencost.search(line):
                field = 'gencost'
            elif bus_name.search(line):
                field = 'bus_name'
            else:
                continue
        elif end.search(line):
            field = None
            continue

        # parse mpc sections
        if field:
            if line.find('=') >= 0:
                line = line.split('=')[1]
            if line.find('[') >= 0:
                line = re.sub(r'\[', '', line)
            elif line.find('{') >= 0:
                line = re.sub(r'{', '', line)

            if line.find('\'') >= 0:  # bus_name
                line = line.split(';')
                data = [i.strip('\'').strip() for i in line]
                mpc['bus_name'].extend(data)
            else:
                if not hasdigit.search(line):
                    continue
                line = line.split(';')
                for item in line:
                    data = [float(val) for val in item.split()]
                    mpc[field].append(data)

    fid.close()

    # add model elements to system
    sw = []

    system.mva = basemva

    for data in mpc['bus']:
        # idx  ty   pd   qd  gs  bs  area  vmag  vang  baseKV  zone  vmax  vmin
        # 0    1    2   3   4   5    6      7     8     9      10    11    12
        idx = int(data[0])
        ty = data[1]
        if ty == 3:
            sw.append(idx)
        pd = data[2] / basemva
        qd = data[3] / basemva
        gs = data[4] / basemva
        bs = data[5] / basemva
        area = data[6]
        vmag = data[7]
        vang = data[8] * deg2rad
        baseKV = data[9]
        zone = data[10]
        vmax = data[11]
        vmin = data[12]

        try:
            system.Bus.elem_add(
                idx=idx,
                name='Bus ' + str(idx),
                Vn=baseKV,
                voltage=vmag,
                angle=vang,
                vmax=vmax,
                vmin=vmin,
                area=area,
                region=zone)
            if pd or qd:
                system.PQ.elem_add(
                    bus=idx, name='PQ ' + str(idx), Vn=baseKV, p=pd, q=qd)
            if gs or bs:
                system.Shunt.elem_add(
                    bus=idx, name='Shunt ' + str(idx), Vn=baseKV, g=gs, b=bs)
        except KeyError:
            logger.error('Error adding <Bus> to powersystem object.')
            retval = False

    gen_idx = 0
    for data in mpc['gen']:
        # bus  pg  qg  qmax  qmin  vg  mbase  status  pmax  pmin  pc1  pc2
        #  0   1   2    3     4     5    6      7       8    9    10    11
        # qc1min  qc1max  qc2min  qc2max  ramp_agc  ramp_10  ramp_30  ramp_q
        #  12      13       14      15      16        17       18      19
        # apf
        #  20

        bus_idx = int(data[0])
        gen_idx += 1
        vg = data[5]
        status = int(data[7])
        mbase = basemva
        pg = data[1] / mbase
        qg = data[2] / mbase
        qmax = data[3] / mbase
        qmin = data[4] / mbase
        pmax = data[8] / mbase
        pmin = data[9] / mbase

        try:
            vn = system.Bus.Vn[system.Bus.uid[bus_idx]]
            if bus_idx in sw:
                system.SW.elem_add(
                    idx=gen_idx,
                    bus=bus_idx,
                    busr=bus_idx,
                    name='SW ' + str(bus_idx),
                    u=status,
                    Vn=vn,
                    v0=vg,
                    pg=pg,
                    qg=qg,
                    pmax=pmax,
                    pmin=pmin,
                    qmax=qmax,
                    qmin=qmin,
                    a0=0.0)
            else:
                system.PV.elem_add(
                    idx=gen_idx,
                    bus=bus_idx,
                    busr=bus_idx,
                    name='PV ' + str(bus_idx),
                    u=status,
                    Vn=vn,
                    v0=vg,
                    pg=pg,
                    qg=qg,
                    pmax=pmax,
                    pmin=pmin,
                    qmax=qmax,
                    qmin=qmin)
        except KeyError:
            logger.error(
                'Error adding <SW> or <PV> to powersystem object.')
            retval = False

    for data in mpc['branch']:
        # fbus	tbus	r	x	b	rateA	rateB	rateC	ratio	angle
        #  0     1      2   3   4     5       6       7       8      9
        # status	angmin	angmax	Pf	Qf	Pt	Qt
        #   10       11       12    13  14  15  16
        fbus = data[0]
        tbus = data[1]
        r = data[2]
        x = data[3]
        b = data[4]
        status = int(data[10])

        if data[8] == 0.0:  # not a transformer
            tf = False
            ratio = 1
            angle = 0
        elif data[8] == 1.0 and data[9] == 0.0:  # not a transformer
            tf = False
            ratio = 1
            angle = 0
        else:
            tf = True
            ratio = data[8]
            angle = data[9]
        try:
            vf = system.Bus.Vn[system.Bus.uid[fbus]]
            vt = system.Bus.Vn[system.Bus.uid[tbus]]
            system.Line.elem_add(
                Vn=vf,
                Vn2=vt,
                bus1=fbus,
                bus2=tbus,
                r=r,
                x=x,
                b=b,
                u=status,
                trasf=tf,
                tap=ratio,
                phi=angle)
        except KeyError:
            logger.error('Error adding <Line> to powersystem object.')
            retval = False
    if len(mpc['bus_name']) == len(system.Bus.name):
        system.Bus.name[:] = mpc['bus_name']

    return retval



import logging

logger = logging.getLogger(__name__)

SHOW_PF_CALL = False
SHOW_INT_CALL = False

all_calls = [
    'gcall', 'gycall', 'fcall', 'fxcall', 'init0', 'pflow', 'windup', 'jac0',
    'init1', 'shunt', 'series', 'flows', 'connection', 'times', 'stagen',
    'dyngen', 'gmcall', 'fmcall', 'dcseries', 'opf', 'obj'
]


class Call(object):
    """ Equation call mamager class for andes routines"""

    def __init__(self, system):
        self.system = system
        self.ndevice = 0
        self.devices = []
        call_strings = [
            'gcalls',
            'fcalls',
            'gycalls',
            'fxcalls',
            'jac0s',
        ]

        self.gisland = 'system.Bus.gisland(system.dae)\n'
        self.gyisland = 'system.Bus.gyisland(system.dae)\n'

        for item in all_calls + call_strings:
            self.__dict__[item] = []

    def setup(self):
        """
        setup the call list after case file is parsed and jit models are loaded
        """
        self.devices = self.system.devman.devices
        self.ndevice = len(self.devices)

        self.gcalls = [''] * self.ndevice
        self.fcalls = [''] * self.ndevice
        self.gycalls = [''] * self.ndevice
        self.fxcalls = [''] * self.ndevice
        self.jac0s = [''] * self.ndevice

        self.build_vec()
        self.build_strings()
        self._compile_newton()
        self._compile_fdpf()
        self._compile_pfload()
        self._compile_pfgen()
        self._compile_seriesflow()
        self._compile_int()
        self._compile_int_f()
        self._compile_int_g()
        self._compile_bus_injection()

    def build_vec(self):
        """build call validity vector for each device"""
        for item in all_calls:
            self.__dict__[item] = []

        for dev in self.devices:
            for item in all_calls:
                if self.system.__dict__[dev].n == 0:
                    val = False
                else:
                    val = self.system.__dict__[dev].calls.get(item, False)
                self.__dict__[item].append(val)

    def build_strings(self):
        """build call string for each device"""
        for idx, dev in enumerate(self.devices):
            header = 'system.' + dev
            self.gcalls[idx] = header + '.gcall(system.dae)\n'
            self.fcalls[idx] = header + '.fcall(system.dae)\n'
            self.gycalls[idx] = header + '.gycall(system.dae)\n'
            self.fxcalls[idx] = header + '.fxcall(system.dae)\n'
            self.jac0s[idx] = header + '.jac0(system.dae)\n'

    def _compile_newton(self):
        """Newton power flow execution
                1. evaluate g and f;
                1.1. handle islanded buses by Bus.gisland()
                2. factorize when needed;
                3. evaluate Gy and Fx.
                3.1. take care of islanded buses by Bus.gyisland()
        """
        string = '"""\n'

        # evaluate algebraic equations g and differential equations f
        string += 'system.dae.init_fg()\n'
        for pflow, gcall, call in zip(self.pflow, self.gcall, self.gcalls):
            if pflow and gcall:
                string += call
        string += 'system.dae.reset_small_g()\n'
        string += '\n'
        for pflow, fcall, call in zip(self.pflow, self.fcall, self.fcalls):
            if pflow and fcall:
                string += call
        string += 'system.dae.reset_small_f()\n'

        # handle islanded buses in algebraic equations
        string += self.gisland
        string += '\n'

        # rebuild constant Jacobian elements if factorization needed
        string += 'if system.dae.factorize:\n'
        string += '    system.dae.init_jac0()\n'
        for pflow, jac0, call in zip(self.pflow, self.jac0, self.jac0s):
            if pflow and jac0:
                string += '    ' + call
        string += '    system.dae.temp_to_spmatrix(\'jac0\')\n'

        # evaluate Jacobians Gy and Fx
        string += 'system.dae.setup_FxGy()\n'
        for pflow, gycall, call in zip(self.pflow, self.gycall, self.gycalls):
            if pflow and gycall:
                string += call
        for pflow, fxcall, call in zip(self.pflow, self.fxcall, self.fxcalls):
            if pflow and fxcall:
                string += call

        # handle islanded buses in the Jacobian
        string += self.gyisland
        string += 'system.dae.temp_to_spmatrix(\'jac\')\n'

        string += '"""'
        if SHOW_PF_CALL:
            logger.debug(string)
        self.newton = compile(eval(string), '', 'exec')

    def _compile_fdpf(self):
        """Fast Decoupled Power Flow execution: Implement g(y)
        """
        string = '"""\n'
        string += 'system.dae.init_g()\n'
        for pflow, gcall, call in zip(self.pflow, self.gcall, self.gcalls):
            if pflow and gcall:
                string += call
        string += 'system.dae.reset_small_g()\n'
        string += '\n'
        string += '"""'
        self.fdpf = compile(eval(string), '', 'exec')

    def _compile_pfload(self):
        """Post power flow computation for load
                  S_gen  + S_line + [S_shunt  - S_load] = 0
        """
        string = '"""\n'
        string += 'system.dae.init_g()\n'
        for gcall, pflow, shunt, stagen, call in zip(
                self.gcall, self.pflow, self.shunt, self.stagen, self.gcalls):
            if gcall and pflow and shunt and not stagen:
                string += call
        string += '\n'
        string += 'system.dae.reset_small_g()\n'
        string += '"""'
        self.pfload = compile(eval(string), '', 'exec')

    def _compile_pfgen(self):
        """Post power flow computation for PV and SW"""
        string = '"""\n'
        string += 'system.dae.init_g()\n'
        for gcall, pflow, shunt, series, stagen, call in zip(
                self.gcall, self.pflow, self.shunt, self.series, self.stagen,
                self.gcalls):
            if gcall and pflow and (shunt or series) and not stagen:
                string += call
        string += '\n'
        string += 'system.dae.reset_small_g()\n'
        string += '"""'
        self.pfgen = compile(eval(string), '', 'exec')

    def _compile_bus_injection(self):
        """Impose injections on buses"""
        string = '"""\n'
        for device, series in zip(self.devices, self.series):
            if series:
                string += 'system.' + device + '.gcall(system.dae)\n'
        string += '\n'
        string += 'system.dae.reset_small_g()\n'
        string += self.gisland
        string += '"""'
        self.bus_injection = compile(eval(string), '', 'exec')

    def _compile_seriesflow(self):
        """Post power flow computation of series device flow"""
        string = '"""\n'
        for device, pflow, series in zip(self.devices, self.pflow,
                                         self.series):
            if pflow and series:
                string += 'system.' + device + '.seriesflow(system.dae)\n'
        string += '\n'
        string += '"""'
        self.seriesflow = compile(eval(string), '', 'exec')

    def _compile_int(self):
        """Time Domain Simulation routine execution"""
        string = '"""\n'

        # evaluate the algebraic equations g
        string += 'system.dae.init_fg(resetz=False)\n'
        for gcall, call in zip(self.gcall, self.gcalls):
            if gcall:
                string += call
        string += '\n'
        string += 'system.dae.reset_small_g()\n'

        # handle islands
        string += self.gisland

        # evaluate differential equations f
        for fcall, call in zip(self.fcall, self.fcalls):
            if fcall:
                string += call
        string += 'system.dae.reset_small_f()\n'
        string += '\n'

        fg_string = string + '"""'
        self.int_fg = compile(eval(fg_string), '', 'exec')

        # rebuild constant Jacobian elements if needed
        string += 'if system.dae.factorize:\n'
        string += '    system.dae.init_jac0()\n'
        for jac0, call in zip(self.jac0, self.jac0s):
            if jac0:
                string += '    ' + call
        string += '    system.dae.temp_to_spmatrix(\'jac0\')\n'

        # evaluate Jacobians Gy and Fx
        string += 'system.dae.setup_FxGy()\n'
        for gycall, call in zip(self.gycall, self.gycalls):
            if gycall:
                string += call
        string += '\n'
        for fxcall, call in zip(self.fxcall, self.fxcalls):
            if fxcall:
                string += call
        string += self.gyisland
        string += 'system.dae.temp_to_spmatrix(\'jac\')\n'

        string += '"""'
        if SHOW_INT_CALL:
            logger.debug(string)

        self.int = compile(eval(string), '', 'exec')

    def _compile_int_f(self):
        """Time Domain Simulation - update differential equations"""
        string = '"""\n'
        string += 'system.dae.init_f()\n'

        # evaluate differential equations f
        for fcall, call in zip(self.fcall, self.fcalls):
            if fcall:
                string += call
        string += 'system.dae.reset_small_f()\n'
        string += '"""'
        self.int_f = compile(eval(string), '', 'exec')

    def _compile_int_g(self):
        """Time Domain Simulation - update algebraic equations and Jacobian"""
        string = '"""\n'

        # evaluate the algebraic equations g
        string += 'system.dae.init_g()\n'
        for gcall, call in zip(self.gcall, self.gcalls):
            if gcall:
                string += call
        string += '\n'
        string += 'system.dae.reset_small_g()\n'

        # handle islands
        string += self.gisland

        # rebuild constant Jacobian elements if needed
        string += 'if system.dae.factorize:\n'
        string += '    system.dae.init_jac0()\n'
        for jac0, call in zip(self.jac0, self.jac0s):
            if jac0:
                string += '    ' + call
        string += '    system.dae.temp_to_spmatrix(\'jac0\')\n'

        # evaluate Jacobians Gy
        string += 'system.dae.setup_Gy()\n'
        for gycall, call in zip(self.gycall, self.gycalls):
            if gycall:
                string += call
        string += '\n'
        string += self.gyisland
        string += 'system.dae.temp_to_spmatrix(\'jac\')\n'

        string += '"""'
        self.int_g = compile(eval(string), '', 'exec')



import numpy as np

from cvxopt import matrix
from cvxopt import mul, exp


def altb(a, b):
    """Return a matrix of logic comparison of A<B"""
    return matrix(np.less(a, b).astype('float'))


def mmax(a, b):
    """Return a matrix of maximum values in a and b element-wise"""
    return matrix(np.maximum(a, b))


def mmin(a, b):
    """Return a matrix of minimum values in a and b element-wise"""
    return matrix(np.minimum(a, b))


def agtb(a, b):
    """Return a matrix of logic comparision of A>B"""
    return matrix(np.greater(a, b).astype('float'))


def aleb(a, b):
    """Return a matrix of logic comparison of A<=B"""
    return matrix(np.less_equal(a, b).astype('float'))


def ageb(a, b):
    """Return a matrix of logic comparision of A>=B"""
    return matrix(np.greater_equal(a, b).astype('float'))


def aeqb(a, b):
    """Return a matrix of logic comparison of A == B"""
    return matrix(np.equal(a, b).astype('float'))


def aneb(a, b):
    """Return a matrix of logic comparison of A != B"""
    return matrix(np.not_equal(a, b).astype('float'))


def aorb(a, b):
    """Return a matrix of logic comparison of A or B"""
    return matrix(np.logical_or(a, b).astype('float'), a.size)


def aandb(a, b):
    """Return a matrix of logic comparison of A or B"""
    return matrix(np.logical_and(a, b).astype('float'), a.size)


def nota(a):
    """Return a matrix of logic negative of A"""
    return matrix(np.logical_not(a).astype('float'), a.size)


def polar(m, a):
    """Return complex number from polar form m*exp(1j*a)"""
    return mul(m, exp(1j * a))


def conj(a):
    """return the conjugate of a"""
    return a.H.T


def neg(u):
    """Return the negative of binary states u"""
    return 1 - u


def mfloor(a):
    """Return the element-wise floor value of a"""
    return matrix(np.floor(a), a.size)


def mround(a):
    """Return the element-wise round value of a"""
    return matrix(np.round(a), a.size)


def not0(a):
    """Return u if u!= 0, return 1 if u == 0"""
    return matrix(list(map(lambda x: 1 if x == 0 else x, a)), a.size)


def zeros(m, n):
    """Return a m-by-n zero-value matrix"""
    return matrix(0.0, (m, n), 'd')


def ones(m, n):
    """Return a m-by-n one-value matrix"""
    return matrix(1.0, (m, n), 'd')


def sign(a):
    """Return the sign of a in (1, -1, 0)"""
    return matrix(np.sign(np.array(a)))


def sort(m, reverse=False):
    """Return sorted m (default: ascending order)"""
    ty = type(m)
    if ty == matrix:
        m = list(m)
    m = sorted(m, reverse=reverse)
    if ty == matrix:
        m = matrix(m)
    return m


def sort_idx(m, reverse=False):
    """Return the indices of m in sorted order (default: ascending order)"""
    return sorted(range(len(m)), key=lambda k: m[k], reverse=reverse)


def index(m, val):
    """
    Return the indices of all the ``val`` in ``m``
    """
    mm = np.array(m)
    idx_tuple = np.where(mm == val)
    idx = idx_tuple[0].tolist()

    return idx


def to_number(s):
    """
    Convert a string to a number.
    If not successful, return the string without blanks
    """
    ret = s
    # try converting to float
    try:
        ret = float(s)
    except ValueError:
        ret = ret.strip('\'').strip()

    # try converting to uid
    try:
        ret = int(s)
    except ValueError:
        pass

    # try converting to boolean
    if ret == 'True':
        ret = True
    elif ret == 'False':
        ret = False
    elif ret == 'None':
        ret = None
    return ret


def sdiv(a, b):
    """Safe division: if a == b == 0, sdiv(a, b) == 1"""
    if len(a) != len(b):
        raise ValueError('Argument a and b does not have the same length')
    idx = 0
    ret = matrix(0, (len(a), 1), 'd')
    for m, n in zip(a, b):
        try:
            ret[idx] = m / n
        except ZeroDivisionError:
            ret[idx] = 1
        finally:
            idx += 1
    return ret



from ..models import order, all_models
from numpy import ndarray
from cvxopt import matrix
import logging

logger = logging.getLogger(__name__)


class DevMan(object):
    """
    Device Manager class.
    Maintains the loaded model list, groups and categories

    """

    def __init__(self, system=None):
        """constructor for devman class"""
        self.system = system
        self.devices = []
        self.group = {}

    def register_device(self, dev_name):
        """register a device to the device list"""
        if dev_name not in self.devices:
            self.devices.append(dev_name)
        group_name = self.system.__dict__[dev_name]._group
        if group_name not in self.group.keys():
            self.group[group_name] = {}

    def register_element(self, dev_name, idx=None):
        """
        Register a device element to the group list

        Parameters
        ----------
        dev_name : str
            model name
        idx : str
            element idx

        Returns
        -------
        str
            assigned idx
        """
        if dev_name not in self.devices:
            logger.error(
                'Device {} missing. call add_device before adding elements'.
                format(dev_name))
            return
        group_name = self.system.__dict__[dev_name]._group
        if idx is None:  # "if not idx" will fail for idx==0.0
            idx = dev_name + '_' + str(len(self.group[group_name].keys()))
        self.group[group_name][idx] = dev_name
        return idx

    def sort_device(self):
        """
        Sort device to follow the order of initialization

        :return: None
        """

        self.devices.sort()
        # idx: the indices of order-sensitive models
        # names: an ordered list of order-sensitive models
        idx = []
        names = []
        for dev in order:
            # if ``dev`` in ``order`` is a model file name:
            #   initialize the models in alphabet order
            if dev in all_models:
                all_dev = list(sorted(all_models[dev].keys()))
                for item in all_dev:
                    if item in self.devices:
                        idx.append(self.devices.index(item))
                        names.append(item)

            # if ``dev`` presents as a model name
            elif dev in self.devices:
                idx.append(self.devices.index(dev))
                names.append(dev)

        idx = sorted(idx)
        for id, name in zip(idx, names):
            self.devices[id] = name

    def get_param(self, group, param, fkey):
        ret = []
        ret_list = False
        if type(fkey) == matrix:
            fkey = list(fkey)
        elif type(fkey) == ndarray:
            fkey = fkey.tolist()

        for key, item in self.group.items():
            if key != group:
                continue
            if type(fkey) != list:
                fkey = [fkey]
            else:
                ret_list = True

            for k in fkey:
                for name, dev in item.items():
                    if name == k:
                        int_id = self.system.__dict__[dev].uid[name]
                        ret.append(
                            self.system.__dict__[dev].__dict__[param][int_id])
                        continue
            if not ret_list:
                ret = ret[0]

        return ret



"""Wind power classes"""
from numpy.random import weibull, uniform  # NOQA
from numpy import arange, log

from cvxopt import matrix, mul, spmatrix, div  # NOQA
from .base import ModelBase

from ..consts import Fx0, Fy0, Gx0, Gy0  # NOQA
from ..consts import Fx, Fy, Gx, Gy  # NOQA
from math import ceil


class WindBase(ModelBase):
    """Base class for wind time series"""

    def __init__(self, system, name):
        super(WindBase, self).__init__(system, name)
        self._group = 'Wind'
        self.param_remove('Sn')
        self.param_remove('Vn')
        self._data.update({
            'T': 1,
            'Vwn': 13,
            'dt': 0.1,
            'rho': 1.225,
        })
        self._descr.update({
            'T': 'Filter time constant',
            'Vwn': 'Wind speed base',
            'dt': 'Sampling time step',
            'rho': 'Air density',
        })
        self._units.update({
            'T': 's',
            'Vwn': 'm/s',
            'rho': 'kg/m^3',
            'dt': 's',
        })
        self._params.extend(['T', 'Vwn', 'dt', 'rho'])
        self._states.extend(['vw'])
        self._fnamex.extend(['v_w'])
        self._algebs.extend(['ws'])
        self._fnamey.extend(['w_s'])
        self._zeros.extend(['T', 'dt'])
        self.time, self.speed, self.t0, self.tf = list(), list(), 0, 0
        self.calls.update({
            'init1': True,
            'gcall': True,
            'fcall': True,
            'jac0': True,
        })

    def setup(self):
        super(WindBase, self).setup()
        # self.vwa = ones(self.n, 1)
        #   todo: remove this after wind turbine _init
        # self.system.dae.x[self.vw] = ones(self.n, 1)

    def servcall(self, dae):
        self.iT = div(1, self.T)
        self.t0 = self.system.tds.config.t0
        self.tf = self.system.tds.config.tf

    def init1(self, dae):
        self.servcall(dae)
        self.time = [0] * self.n
        self.speed = [0] * self.n
        for i in range(self.n):
            self.time[i] = list(
                arange(self.t0, self.tf + self.dt[i], self.dt[i]))
            self.speed[i] = [0] * len(self.time[i])

        self.generate(dae)

        dae.y[self.ws] = dae.x[self.vw]
        self.speed[:][0] = dae.y[self.ws]

    def generate(self, dae):
        """Generate the wind speed time and data points"""
        pass

    def windspeed(self, t):
        """Return the wind speed list at time `t`"""
        ws = [0] * self.n

        for i in range(self.n):
            q = ceil(t / self.dt[i])
            q_prev = 0 if q == 0 else q - 1

            r = t % self.dt[i]
            r = 0 if abs(r) < 1e-6 else r

            if r == 0:
                ws[i] = self.speed[i][q]
            else:
                t1 = self.time[i][q_prev]
                s1 = self.speed[i][q_prev]
                s2 = self.speed[i][q]
                ws[i] = s1 + (t - t1) * (s2 - s1) / self.dt[i]

        return matrix(ws)

    def gcall(self, dae):
        dae.g[self.ws] = self.windspeed(dae.t) - dae.y[self.ws]

    def fcall(self, dae):
        dae.f[self.vw] = mul(dae.y[self.ws] - dae.x[self.vw], self.iT)

    def jac0(self, dae):
        dae.add_jac(Gy0, -1, self.ws, self.ws)
        dae.add_jac(Fx0, -self.iT, self.vw, self.vw)
        dae.add_jac(Fy0, self.iT, self.vw, self.ws)


class Weibull(WindBase):
    """Weibull distribution wind speed class"""

    def __init__(self, system, name):
        super(Weibull, self).__init__(system, name)
        self._name = 'Weibull'
        self._data.update({'c': 5.0, 's': 2.0})
        self._descr.update({
            'c': 'Scale factor',
            's': 'Shape factor',
        })
        self._zeros.extend(['c', 's'])
        self._params.extend(['c', 's'])
        self._init()

    def generate(self, dae):
        for i in range(self.n):
            npoint = len(self.time[i])
            if self.c[i] <= 0.0:
                self.c[i] = 5.0
            if self.s[i] <= 0.0:
                self.s[i] = 2.0
            sample = (-log(uniform(0, 1,
                                   (npoint, 1))) / self.c[i])**(1 / self.s[i])

            avg = sum(sample) / npoint
            sample[0] = dae.x[self.vw[i]]
            sample[1:] = abs(sample[1:] - avg + 1) * sample[0]

            self.speed[i] = matrix(sample)


class ConstWind(WindBase):
    """Constant wind power class"""

    def __init__(self, system, name):
        super(ConstWind, self).__init__(system, name)
        self._name = 'ConstWind'
        self._init()

    def generate(self, dae):
        for i in range(self.n):
            sample = [dae.x[self.vw[i]]] * len(self.time[i])
            self.speed[i] = list(sample)



"""Control blocks"""
from cvxopt import matrix, sparse, spmatrix  ## NOQA
from cvxopt import mul, div, log, sin, cos  ## NOQA
from .base import ModelBase  ## NOQA
# from ..consts import *
# from ..utils.math import *


class PI1(object):
    """PI controller class as addon base class"""

    def __init__(self, params=None, inputs=None, outputs=None):

        assert type(outputs) == list

        if hasattr(self, 'nPI1'):
            self.nPI1 += 1
        else:
            self.nPI1 = 1

        if hasattr(self, 'PI1_label'):
            self.PI1_label.append('PI1_' + str(self.nPI1))
        else:
            self.PI1_label = ['PI1' + str(self.nPI1)]

        if params:
            assert isinstance(params, dict)
        else:
            params = {
                'Kp' + str(self.nPI1): 0.1,
                'Ki' + str(self.nPI1): 0,
            }

        self._algebs = ['']
        if inputs:
            assert type(inputs) == list
        else:
            inputs = ['PI_IN_']



# ANDES, a power system simulation tool for research.
#
# Copyright 2015-2017 Hantao Cui
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Base class for building ANDES models
"""

import sys
import importlib
import numpy as np
import logging

from cvxopt import matrix, spmatrix  # NOQA
from cvxopt import mul, div

from logging import DEBUG, INFO, WARNING, ERROR, CRITICAL  # NOQA

from ..utils.math import agtb, altb, index, zeros  # NOQA
from ..utils.tab import Tab

logger = logging.getLogger(__name__)

pd = None


class ModelBase(object):
    """base class for power system device models"""

    def __init__(self, system, name):
        """meta-data to be overloaded by subclasses"""

        # metadata list and dictionaries
        self._param_attr_lists = ('_params', '_zeros', '_mandatory', '_powers',
                                  '_voltages', '_currents', '_z', '_y', '_r',
                                  '_g', '_dccurrents', '_dcvoltages', '_times')

        self._param_attr_dicts = ('_data', '_units', '_descr')

        self.system = system
        # self.n = 0    # device count
        self.u = []  # device status
        self.idx = []  # external index list
        self.uid = {}  # mapping from `self.idx` to unique positional indices
        self.name = []  # element name list

        # identifications
        self._name = name
        self._group = None
        self._category = None

        # interfaces
        self._ac = {}  # ac bus variables
        self._dc = {}  # dc node variables
        self._ctrl = {}  # controller interfaces

        # variables
        self._states = []
        self._algebs = []
        self._algebs_intf = []

        self._states_descr = {}
        self._algebs_descr = {}

        # variable names
        self._unamex = []
        self._unamey = []
        self._fnamex = []
        self._fnamey = []

        # `self.mdl_from`:
        #    a list of lists, which contain tuples of (model, idx)
        #    of elements connected to the corresponding element
        # `self.mdl_to`:
        #    a list of lists, which contain tuples of (model, idx) of
        #    elements the corresponding element connects to

        self.mdl_from = []
        self.mdl_to = []

        # parameters to be converted to matrix
        self._params = ['u', 'Sn', 'Vn']

        # default parameter data
        self._data = {
            'u': 1,
            'Sn': 100.0,
            'Vn': 110.0,
        }

        # units of parameters
        self._units = {
            'u': 'bool',
            'Sn': 'MVA',
            'Vn': 'kV',
        }

        # variable descriptions
        self._descr = {
            'u': 'Connection status',
            'Sn': 'Power rating',
            'Vn': 'Ac Voltage rating',
        }
        # non-zero parameters
        self._zeros = ['Sn', 'Vn']

        # mandatory variables
        self._mandatory = []

        # service/temporary variables
        self._service = []
        self._service_ty = []

        self._store = {}
        self._snapshot = {}

        # equation storage
        self._equations = []

        # parameters to be per-unitized
        self._powers = []  # powers, inertia and damping
        self._voltages = []  # ac voltages
        self._currents = []  # ac currents
        self._z = []  # ac impedance
        self._y = []  # ac admittance

        self._dccurrents = []  # dc currents
        self._dcvoltages = []  # dc voltages
        self._r = []  # dc resistance
        self._g = []  # dc susceptance

        self._times = []  # time constants

        self._event_times = []  # event occurrance times

        # property functions this device has

        self.calls = dict(
            pflow=False,
            addr1=False,
            init0=False,
            init1=False,
            jac0=False,
            windup=False,
            gcall=False,
            fcall=False,
            gycall=False,
            fxcall=False,
            series=False,
            shunt=False,
            flows=False,
            dcseries=False,
        )

        self._flags = {
            'sysbase': False,
            'allocate': False,
            'address': False,
        }

        self._config = {
            'address_group_by': 'element',
            'is_series': False,
        }

        # pandas.DataFrame
        self.param_dict = {}
        self.param_df = {}

        self.define()

    def define(self):
        """
        Hook function where derived models define parameters, variables,
        service constants, and equations

        :return:
        """
        # raise NotImplemented('Subclasses must overwrite this method')
        pass

    def _init(self):
        """
        Convert model metadata to class attributes.

        This function is called automatically after ``define()`` in new
        versions.

        :return: None
        """
        assert self._name
        assert self._group

        # self.n = 0
        self.u = []
        self.name = []
        self.idx = []
        self.uid = {}

        if not self._unamey:
            self._unamey = self._algebs
        else:
            assert len(self._unamey) == len(self._algebs)

        if not self._unamex:
            self._unamex = self._states
        else:
            assert len(self._unamex) == len(self._states)

        for item in self._data.keys():
            self.__dict__[item] = []

        for bus in self._ac.keys():
            for var in self._ac[bus]:
                self.__dict__[var] = []
        for node in self._dc.keys():
            for var in self._dc[node]:
                self.__dict__[var] = []

        for var in self._states + self._algebs + self._service:
            self.__dict__[var] = []

        self._flags['sysbase'] = False
        self._flags['allocate'] = False
        self._flags['address'] = False

    def param_define(self,
                     param,
                     default,
                     unit='',
                     descr='',
                     tomatrix=True,
                     nonzero=False,
                     mandatory=False,
                     power=False,
                     voltage=False,
                     current=False,
                     z=False,
                     y=False,
                     r=False,
                     g=False,
                     dccurrent=False,
                     dcvoltage=False,
                     time=False,
                     event_time=False,
                     **kwargs):
        """
        Define a parameter in the model

        :param tomatrix: convert this parameter list to matrix
        :param param: parameter name
        :param default: parameter default value
        :param unit: parameter unit
        :param descr: description
        :param nonzero: is non-zero
        :param mandatory: is mandatory
        :param power: is a power value in the `self.Sn` base
        :param voltage: is a voltage value in the `self.Vn` base
        :param current: is a current value in the device base
        :param z: is an impedance value in the device base
        :param y: is an admittance value in the device base
        :param r: is a dc resistance value in the device base
        :param g: is a dc conductance value in the device base
        :param dccurrent: is a dc current value in the device base
        :param dcvoltage: is a dc votlage value in the device base
        :param time: is a time value in the device base
        :param event_time: is a variable for timed event

        :type param: str
        :type tomatrix: bool
        :type default: str, float
        :type unit: str
        :type descr: str
        :type nonzero: bool
        :type mandatory: bool
        :type power: bool
        :type voltage: bool
        :type current: bool
        :type z: bool
        :type y: bool
        :type r: bool
        :type g: bool
        :type dccurrent: bool
        :type dcvoltage: bool
        :type time: bool
        :type event_time: bool
        """
        assert param not in self._data
        assert param not in self._algebs
        assert param not in self._states
        assert param not in self._service

        self._data.update({param: default})
        if unit:
            self._units.update({param: unit})
        if descr:
            self._descr.update({param: descr})
        if tomatrix:
            self._params.append(param)
        if nonzero:
            self._zeros.append(param)
        if mandatory:
            self._mandatory.append(param)
        if power:
            self._powers.append(param)
        if voltage:
            self._voltages.append(param)
        if current:
            self._currents.append(param)
        if z:
            self._z.append(param)
        if y:
            self._y.append(param)
        if r:
            self._r.append(param)
        if g:
            self._g.append(param)
        if dccurrent:
            self._dccurrents.append(param)
        if dcvoltage:
            self._dcvoltages.append(param)
        if time:
            self._times.append(param)
        if event_time:
            self._event_times.append(param)

    def var_define(self, variable, ty, fname, descr='', uname=''):
        """
        Define a variable in the model

        :param fname: LaTex formatted variable name string
        :param uname: unformatted variable name string, `variable` as default
        :param variable: variable name
        :param ty: type code in ``('x', 'y')``
        :param descr: variable description

        :type variable: str
        :type ty: str
        :type descr: str
        :return:
        """
        assert ty in ('x', 'y')
        if not uname:
            uname = variable

        if ty == 'x':
            self._states.append(variable)
            self._fnamex.append(fname)
            self._unamex.append(uname)
            if descr:
                self._states_descr.update({variable: descr})
        elif ty == 'y':
            self._algebs.append(variable)
            self._fnamey.append(fname)
            self._unamey.append(uname)
            if descr:
                self._algebs_descr.update({variable: descr})

    def service_define(self, service, ty):
        """
        Add a service variable of type ``ty`` to this model

        :param str service: variable name
        :param type ty: variable type
        :return: None
        """

        assert service not in self._data
        assert service not in self._algebs + self._states

        self._service.append(service)
        self._service_ty.append(ty)

    def get_uid(self, idx):
        """
        Return the `uid` of the elements with the given `idx`

        :param list, matrix idx: external indices
        :type idx: list, matrix
        :return: a matrix of uid
        """
        assert idx is not None

        if isinstance(idx, (int, float, str)):
            return self.uid[idx]

        ret = []
        for i in idx:
            tmp = self.uid.get(i, None)
            assert tmp is not None, (
                'Model <{}> does not have element <{}>'.format(self._name, i))
            ret.append(self.uid[i])

        return ret

    def get_idx(self, uid):
        """
        Return the ``idx`` of the elements whose internal indices are ``uid``

        :param uid: uid of elements to query
        :return: idx of the elements
        """
        return [self.idx[i] for i in uid]

    def get_field(self, field, idx=None, astype=None):
        """
        Return `self.field` for the elements labeled by `idx`

        :param astype: type cast of the return value
        :param field: field name of this model
        :param idx: element indices, will be the whole list if not specified
        :return: field values
        """
        assert astype in (None, list, matrix)
        ret = None

        if idx is None:
            idx = self.idx

        # ===================disable warning ==============================
        # if field in self._service:
        # logger.warning(
        #     'Reading service variable <{model}.{field}> could be unsafe.'
        #         .format(field=field, model=self._name)
        # )
        # =================================================================

        uid = self.get_uid(idx)

        field_data = self.__dict__[field]

        if isinstance(field_data, matrix):
            ret = field_data[uid]
        elif isinstance(field_data, list):
            if isinstance(idx, (float, int, str)):
                ret = field_data[uid]
            else:
                ret = [field_data[x] for x in uid]

        if astype is not None:
            ret = astype(ret)

        return ret

    def _alloc(self):
        """
        Allocate empty memory for dae variable indices.
        Called in device setup phase.

        :return: None
        """
        nzeros = [0] * self.n
        for var in self._states:
            self.__dict__[var] = nzeros[:]
        for var in self._algebs:
            self.__dict__[var] = nzeros[:]

    def data_to_dict(self, sysbase=False):
        """
        Return the loaded model parameters as one dictionary.

        Each key of the dictionary is a parameter name, and the value is a
        list of all the parameter values.

        :param sysbase: use system base quantities
        :type sysbase: bool
        """
        assert isinstance(sysbase, bool)

        ret = {}

        for key in self.data_keys:
            if (not sysbase) and (key in self._store):
                val = self._store[key]
            else:
                val = self.__dict__[key]

            ret[key] = val

        return ret

    @property
    def n(self):
        """
        Return the count of elements

        Returns
        -------
        int:
            count of elements
        """

        return len(self.idx)

    @property
    def data_keys(self):
        """
        Returns a list of all parameters plus ``name`` and ``idx``

        Returns
        -------
        list of parameters
        """

        keys = ['idx', 'name', 'u'] + sorted(self._data.keys())
        return set(keys)

    def data_to_list(self, sysbase=False):
        """
        Return the loaded model data as a list of dictionaries.

        Each dictionary contains the full parameters of an element.
        :param sysbase: use system base quantities
        :type sysbase: bool
        """
        ret = list()

        # for each element
        for i in range(self.n):
            # read the parameter values and put in the temp dict ``e``
            e = {}
            for key in self.data_keys:
                if sysbase and (key in self._store):
                    val = self._store[key][i]
                else:
                    val = self.__dict__[key][i]
                e[key] = val

            ret.append(e)

        return ret

    def data_from_list(self, data):
        """
        Populate model parameters from a list of parameter dictionaries

        Parameters
        ----------
        data : list
            List of parameter dictionaries

        Returns
        -------
        None
        """
        for item in data:
            self.elem_add(**item)

    def data_from_dict(self, data):
        """
        Populate model parameters from a dictionary of parameters

        Parameters
        ----------
        data : dict
            List of parameter dictionaries

        Returns
        -------
        None
        """

        nvars = []
        for key, val in data.items():
            self.__dict__[key].extend(val)

            # assure the same parameter matrix size
            if len(nvars) > 1 and len(val) != nvars[-1]:
                raise IndexError(
                    'Model <{}> parameter <{}> must have the same length'.
                    format(self._name, key))
            nvars.append(len(val))

        # assign idx-uid mapping
        for i, idx in zip(range(self.n), self.idx):
            self.uid[idx] = i

    def data_to_df(self, sysbase=False):
        """
        Return a pandas.DataFrame of device parameters.
        :param sysbase: save per unit values in system base
        """

        p_dict_comp = self.data_to_dict(sysbase=sysbase)
        self._check_pd()

        self.param_df = pd.DataFrame(data=p_dict_comp).set_index('idx')

        return self.param_df

    def var_to_df(self):
        """
        Return the current var_to_df of variables

        :return: pandas.DataFrame
        """
        ret = {}
        self._check_pd()

        if self._flags['address'] is False:
            return pd.DataFrame.from_dict(ret)

        ret.update({'name': self.name})
        ret.update({'idx': self.idx})

        for x in self._states:
            idx = self.__dict__[x]
            ret.update({x: self.system.dae.x[idx]})
        for y in self._algebs:
            idx = self.__dict__[y]
            ret.update({y: self.system.dae.y[idx]})

        var_df = pd.DataFrame.from_dict(ret).set_index('idx')

        return var_df

    @staticmethod
    def _check_pd():
        """
        Import pandas to globals() if not exist
        """
        if globals()['pd'] is None:
            globals()['pd'] = importlib.import_module('pandas')

        return

    def param_remove(self, param: 'str') -> None:
        """
        Remove a param from this model

        :param param: name of the parameter to be removed
        :type param: str
        """
        for attr in self._param_attr_dicts:
            if param in self.__dict__[attr]:
                self.__dict__[attr].pop(param)

        for attr in self._param_attr_lists:
            if param in self.__dict__[attr]:
                self.__dict__[attr].remove(param)

    def param_alter(self,
                    param,
                    default=None,
                    unit=None,
                    descr=None,
                    tomatrix=None,
                    nonzero=None,
                    mandatory=None,
                    power=None,
                    voltage=None,
                    current=None,
                    z=None,
                    y=None,
                    r=None,
                    g=None,
                    dccurrent=None,
                    dcvoltage=None,
                    time=None,
                    **kwargs):
        """
        Set attribute of an existing parameter.
        To be used to alter an attribute inherited from parent models.

        See .. self.param_define for argument descriptions.
        """
        assert param in self._data, \
            'parameter <{}> does not exist in {}'.format(param, self._name)

        def alter_attr(p, attr, value):
            """Set self.__dict__[attr] for param based on value
            """
            if value is None:
                return
            elif (value is True) and (p not in self.__dict__[attr]):
                self.__dict__[attr].append(p)
            elif (value is False) and (p in self.__dict__[attr]):
                self.__dict__[attr].remove(p)
            else:
                self.log('No need to alter {} for {}'.format(attr, p))

        if default is not None:
            self._data.update({param: default})
        if unit is not None:
            self._units.update({param: unit})
        if descr is not None:
            self._descr.update({param: descr})

        alter_attr(param, '_params', tomatrix)
        alter_attr(param, '_zeros', nonzero)
        alter_attr(param, '_mandatory', mandatory)
        alter_attr(param, '_powers', power)
        alter_attr(param, '_voltages', voltage)
        alter_attr(param, '_currents', current)
        alter_attr(param, '_z', z)
        alter_attr(param, '_y', y)
        alter_attr(param, '_r', r)
        alter_attr(param, '_g', g)
        alter_attr(param, '_dcvoltages', dcvoltage)
        alter_attr(param, '_dccurrents', dccurrent)
        alter_attr(param, '_times', time)

    def eq_add(self, expr, var, intf=False):
        """
        Add an equation to this model.

        An equation is associated with the addresses of a variable.  The
        number of equations must equal that of variables.

        Stored to ``self._equations`` is a tuple of ``(expr, var, intf,
        ty)`` where ``ty`` is in ('f', 'g')

        :param str expr: equation expression
        :param str var: variable name to be associated with
        :param bool intf: if this equation is added to an interface equation,
            namely, an equation outside this model

        :return: None

        """

        # determine the type of the equation, ('f', 'g') based on var

        # We assume that all differential equations are only modifiable by
        #   the model itself
        # Only interface to algebraic varbailes of external models

        ty = ''
        if var in self._algebs:
            ty = 'g'
        elif var in self._states:
            ty = 'f'
        else:
            for _, item in self._ac:
                if var in item:
                    ty = 'g'
                    if intf is False:
                        intf = True
            for _, item in self._dc:
                if var == item:
                    ty = 'g'
                    if intf is False:
                        intf = True
        if ty == '':
            self.log(
                'Equation associated with interface variable {var} '
                'assumed as algeb'.
                format(var=var),
                DEBUG)
            ty = 'g'

        self._equations.append((expr, var, intf, ty))

    def read_data_ext(self, model: str, field: str, idx=None, astype=None):
        """
        Return a field of a model or group at the given indices

        :param str model: name of the group or model to retrieve
        :param str field: name of the field
        :param list, int, float str idx: idx of elements to access
        :param type astype: type cast
        :return:
        """
        ret = list()

        if model in self.system.devman.devices:
            ret = self.system.__dict__[model].get_field(field, idx)

        elif model in self.system.devman.group.keys():
            # ===============================================================
            # Since ``self.system.devman.group`` is an unordered dictionary,
            #   ``idx`` must be given to retrieve ``field`` across models.
            #
            # Returns a matrix by default
            # ===============================================================
            assert idx is not None, \
                'idx must be specified when accessing group fields'
            astype = matrix

            for item in idx:
                dev_name = self.system.devman.group[model].get(item, None)
                ret.append(self.read_data_ext(dev_name, field, idx=item))

        else:
            raise NameError(
                'Model or Group <{0}> does not exist.'.format(model))

        if (ret is None) or isinstance(ret, (int, float, str)):
            return ret
        elif astype is None:
            return ret
        else:
            return astype(ret)

    def copy_data_ext(self, model, field, dest=None, idx=None, astype=None):
        """
        Retrieve the field of another model and store it as a field.

        :param model: name of the source model being a model name or a group name
        :param field: name of the field to retrieve
        :param dest: name of the destination field in ``self``
        :param idx: idx of elements to access
        :param astype: type cast

        :type model: str
        :type field: str
        :type dest: str
        :type idx: list, matrix
        :type astype: None, list, matrix

        :return: None

        """
        # use default destination

        if not dest:
            dest = field
        assert dest not in self._states + self._algebs

        self.__dict__[dest] = self.read_data_ext(
            model, field, idx, astype=astype)

        if idx is not None:
            if len(idx) == self.n:
                self.link_to(model, idx, self.idx)

    def elem_add(self, idx=None, name=None, **kwargs):
        """
        Add an element of this model

        :param idx: element idx
        :param name: element name
        :param kwargs: keyword arguments of the parameters
        :return: allocated idx
        """

        idx = self.system.devman.register_element(dev_name=self._name, idx=idx)
        self.system.__dict__[self._group].register_element(self._name, idx)

        self.uid[idx] = self.n
        self.idx.append(idx)

        self.mdl_to.append(list())
        self.mdl_from.append(list())

        # self.n += 1

        if name is None:
            self.name.append(self._name + ' ' + str(self.n))
        else:
            self.name.append(name)

        # check mandatory parameters
        for key in self._mandatory:
            if key not in kwargs.keys():
                self.log(
                    'Mandatory parameter <{:s}.{:s}> missing'.format(
                        self.name[-1], key), ERROR)
                sys.exit(1)

        # set default values
        for key, value in self._data.items():
            self.__dict__[key].append(value)

        # overwrite custom values
        for key, value in kwargs.items():
            if key not in self._data:
                self.log(
                    'Parameter <{:s}.{:s}> is not used.'.format(
                        self.name[-1], key), WARNING)
                continue
            self.__dict__[key][-1] = value

            # check data consistency
            if not value and key in self._zeros:
                if key == 'Sn':
                    default = self.system.mva
                elif key == 'fn':
                    default = self.system.config.freq
                else:
                    default = self._data[key]
                self.__dict__[key][-1] = default
                self.log(
                    'Using default value for <{:s}.{:s}>'.format(
                        self.name[-1], key), WARNING)

        return idx

    def elem_remove(self, idx=None):
        """
        Remove elements labeled by idx from this model instance.

        :param list,matrix idx: indices of elements to be removed
        :return: None
        """
        if idx is not None:
            if idx in self.uid:
                key = idx
                item = self.uid[idx]
            else:
                self.log('The item <{:s}> does not exist.'.format(idx), ERROR)
                return None
        else:
            return None

        convert = False
        if isinstance(self.__dict__[self._params[0]], matrix):
            self._param_to_list()
            convert = True

        # self.n -= 1
        self.uid.pop(key, '')
        self.idx.pop(item)
        self.mdl_from.pop(item)
        self.mdl_to.pop(item)

        for x, y in self.uid.items():
            if y > item:
                self.uid[x] = y - 1

        for param in self._data:
            self.__dict__[param].pop(item)

        for param in self._service:
            if len(self.__dict__[param]) == (self.n + 1):
                if isinstance(self.__dict__[param], list):
                    self.__dict__[param].pop(item)
                elif isinstance(self.__dict__[param], matrix):
                    service = list(self.__dict__[param])
                    service.pop(item)
                    self.__dict__[param] = matrix(service)

        for x in self._states:
            if len(self.__dict__[x]):
                self.__dict__[x].pop(item)

        for y in self._algebs:
            if self.__dict__[y]:
                self.__dict__[y].pop(item)

        for key, param in self._ac.items():
            if isinstance(param, list):
                for subparam in param:
                    if len(self.__dict__[subparam]):
                        self.__dict__[subparam].pop(item)
            else:
                self.__dict__[param].pop(item)

        for key, param in self._dc.items():
            self.__dict__[param].pop(item)

        self.name.pop(item)
        if convert and self.n:
            self._param_to_matrix()

    def data_to_sys_base(self):
        """
        Converts parameters to system base. Stores a copy in ``self._store``.
        Sets the flag ``self.flag['sysbase']`` to True.

        :return: None
        """
        if (not self.n) or self._flags['sysbase']:
            return
        Sb = self.system.mva
        Vb = matrix([])
        if 'bus' in self._ac.keys():
            Vb = self.read_data_ext('Bus', 'Vn', idx=self.bus)
        elif 'bus1' in self._ac.keys():
            Vb = self.read_data_ext('Bus', 'Vn', idx=self.bus1)

        for var in self._voltages:
            self._store[var] = self.__dict__[var]
            self.__dict__[var] = mul(self.__dict__[var], self.Vn)
            self.__dict__[var] = div(self.__dict__[var], Vb)
        for var in self._powers:
            self._store[var] = self.__dict__[var]
            self.__dict__[var] = mul(self.__dict__[var], self.Sn)
            self.__dict__[var] /= Sb
        for var in self._currents:
            self._store[var] = self.__dict__[var]
            self.__dict__[var] = mul(self.__dict__[var], self.Sn)
            self.__dict__[var] = div(self.__dict__[var], self.Vn)
            self.__dict__[var] = mul(self.__dict__[var], Vb)
            self.__dict__[var] /= Sb
        if len(self._z) or len(self._y):
            Zn = div(self.Vn**2, self.Sn)
            Zb = (Vb**2) / Sb
            for var in self._z:
                self._store[var] = self.__dict__[var]
                self.__dict__[var] = mul(self.__dict__[var], Zn)
                self.__dict__[var] = div(self.__dict__[var], Zb)
            for var in self._y:
                self._store[var] = self.__dict__[var]
                if self.__dict__[var].typecode == 'd':
                    self.__dict__[var] = div(self.__dict__[var], Zn)
                    self.__dict__[var] = mul(self.__dict__[var], Zb)
                elif self.__dict__[var].typecode == 'z':
                    self.__dict__[var] = div(self.__dict__[var], Zn + 0j)
                    self.__dict__[var] = mul(self.__dict__[var], Zb + 0j)

        if len(self._dcvoltages) or len(self._dccurrents) or len(
                self._r) or len(self._g):
            dckey = sorted(self._dc.keys())[0]
            Vbdc = self.read_data_ext('Node', 'Vdcn', self.__dict__[dckey])
            Ib = div(Sb, Vbdc)
            Rb = div(Vbdc, Ib)

        for var in self._dcvoltages:
            self._store[var] = self.__dict__[var]
            self.__dict__[var] = mul(self.__dict__[var], self.Vdcn)
            self.__dict__[var] = div(self.__dict__[var], Vbdc)

        for var in self._dccurrents:
            self._store[var] = self.__dict__[var]
            self.__dict__[var] = mul(self.__dict__[var], self.Idcn)
            self.__dict__[var] = div(self.__dict__[var], Ib)

        for var in self._r:
            self._store[var] = self.__dict__[var]
            self.__dict__[var] = div(self.__dict__[var], Rb)

        for var in self._g:
            self._store[var] = self.__dict__[var]
            self.__dict__[var] = mul(self.__dict__[var], Rb)

        self._flags['sysbase'] = True

    def data_to_elem_base(self):
        """
        Convert parameter data to element base

        Returns
        -------
        None
        """
        if self._flags['sysbase'] is False:
            return

        for key, val in self._store:
            self.__dict__[key] = val

        self._flags['sysbase'] = False

    def setup(self):
        """
        Set up empty class structure and allocate empty memory
        for variable addresses.

        :return: None
        """
        if not self.n:
            return

        self._param_to_matrix()
        self._alloc()

    def _intf_network(self):
        """
        Retrieve the ac and dc network interface variable indices.

        :Example:

            ``self._ac = {'bus1': (a1, v1)}`` gives
             - indices: self.bus1
             - system.Bus.a -> self.a1
             - system.Bus.v -> self.v1

            ``self._dc = {'node1': v1}`` gives
              - indices: self.node1
              - system.Node.v-> self.v1

        :return: None
        """

        for key, val in self._ac.items():
            self.copy_data_ext(
                model='Bus', field='a', dest=val[0], idx=self.__dict__[key])
            self.copy_data_ext(
                model='Bus', field='v', dest=val[1], idx=self.__dict__[key])

        for key, val in self._dc.items():
            self.copy_data_ext(
                model='Node', field='v', dest=val, idx=self.__dict__[key])

        # check for interface voltage differences
        self._check_Vn()

    def _intf_ctrl(self):
        """
        Retrieve variable indices of controlled models.

        Control interfaces are specified in ``self._ctrl``.
        Each ``key:value`` pair has ``key`` being the variable names
        for the reference idx and ``value`` being a tuple of
        ``(model name, field to read, destination field, return type)``.

        :Example:

            ``self._ctrl = {'syn', ('Synchronous', 'omega', 'w', list)}``
             - indices: self.syn
             - self.w = list(system.Synchronous.omega)

        :return: None
        """

        for key, val in self._ctrl.items():
            model, field, dest, astype = val
            self.copy_data_ext(
                model, field, dest=dest, idx=self.__dict__[key], astype=astype)

    def _addr(self):
        """
        Assign dae addresses for algebraic and state variables.

        Addresses are stored in ``self.__dict__[var]``.
        ``dae.m`` and ``dae.n`` are updated accordingly.

        Returns
        -------
        None

        """
        group_by = self._config['address_group_by']

        assert not self._flags['address'], "{} address already assigned".format(self._name)
        assert group_by in ('element', 'variable')

        m0 = self.system.dae.m
        n0 = self.system.dae.n
        mend = m0 + len(self._algebs) * self.n
        nend = n0 + len(self._states) * self.n

        if group_by == 'variable':
            for idx, item in enumerate(self._algebs):
                self.__dict__[item] = list(
                    range(m0 + idx * self.n, m0 + (idx + 1) * self.n))
            for idx, item in enumerate(self._states):
                self.__dict__[item] = list(
                    range(n0 + idx * self.n, n0 + (idx + 1) * self.n))
        elif group_by == 'element':
            for idx, item in enumerate(self._algebs):
                self.__dict__[item] = list(
                    range(m0 + idx, mend, len(self._algebs)))
            for idx, item in enumerate(self._states):
                self.__dict__[item] = list(
                    range(n0 + idx, nend, len(self._states)))

        self.system.dae.m = mend
        self.system.dae.n = nend

        self._flags['address'] = True

    def _varname(self):
        """
        Set up variable names in ``self.system.varname``.

        Variable names follows the convention ``VariableName,Model Name``.
        A maximum of 24 characters are allowed for each variable.

        :return: None
        """
        if not self._flags['address']:
            self.log('Unable to assign Varname before allocating address',
                     ERROR)
            return

        varname = self.system.varname
        for i in range(self.n):
            iname = str(self.name[i])

            for e, var in enumerate(self._states):
                unamex = self._unamex[e]
                fnamex = self._fnamex[e]
                idx = self.__dict__[var][i]

                varname.unamex[idx] = '{} {}'.format(unamex, iname)[:33]
                varname.fnamex[idx] = '$' + '{}\\ {}'.format(
                    fnamex, iname.replace(' ', '\\ '))[:33] + '$'

            for e, var in enumerate(self._algebs):
                unamey = self._unamey[e]
                fnamey = self._fnamey[e]
                idx = self.__dict__[var][i]

                varname.unamey[idx] = '{} {}'.format(unamey, iname)[:33]
                varname.fnamey[idx] = '$' + '{}\\ {}'.format(
                    fnamey, iname.replace(' ', '\\ '))[:33] + '$'

    def _param_to_matrix(self):
        """
        Convert parameters defined in `self._params` to `cvxopt.matrix`

        :return None
        """
        for item in self._params:
            self.__dict__[item] = matrix(self.__dict__[item], tc='d')

    def _param_to_list(self):
        """
        Convert parameters defined in `self._param` to list

        :return None
        """
        for item in self._params:
            self.__dict__[item] = list(self.__dict__[item])

    def log(self, msg, level=INFO):
        """Record a line of log in logger

        :param str msg: content of the messag

        :param level: logging level
        :return: None

        """
        logger.log(level, '<{}> - '.format(self._name) + msg)

    def init_limit(self, key, lower=None, upper=None, limit=False):
        """ check if data is within limits. reset if violates"""
        above = agtb(self.__dict__[key], upper)
        for idx, item in enumerate(above):
            if item == 0.:
                continue
            maxval = upper[idx]
            self.log(
                '{0} <{1}.{2}> above its maximum of {3}.'.format(
                    self.name[idx], self._name, key, maxval), ERROR)
            if limit:
                self.__dict__[key][idx] = maxval

        below = altb(self.__dict__[key], lower)
        for idx, item in enumerate(below):
            if item == 0.:
                continue
            minval = lower[idx]
            self.log(
                '{0} <{1}.{2}> below its minimum of {3}.'.format(
                    self.name[idx], self._name, key, minval), ERROR)
            if limit:
                self.__dict__[key][idx] = minval

    def __repr__(self):

        ret = ''
        ret += '\n'
        ret += 'Model <{:s}> parameters in element base\n'.format(self._name)
        ret += self.data_to_df(sysbase=False).to_string()

        ret += '\n'
        ret += 'Model <{:s}> variable snapshot\n'.format(self._name)
        ret += self.var_to_df().to_string()

        return ret

    def __str__(self):
        super(ModelBase, self).__repr__()
        return '<{name}> with {n} elements [{mem}]'.format(
            name=self._name,
            mem=hex(id(self)),
            n=self.n,
        )

    def doc(self, export='plain'):
        """
        Build help document into a Texttable table

        :param ('plain', 'latex') export: export format
        :param save: save to file ``help_model.extension`` or not
        :param writemode: file write mode

        :return: None
        """
        title = '<{}.{}>'.format(self._group, self._name)
        table = Tab(export=export, title=title, descr=self.__doc__)
        rows = []
        keys = sorted(self._data.keys())
        for key in keys:
            val = self._data[key]
            suf = ''
            if key in self._mandatory:
                suf = ' *'

            elif key in self._powers + \
                    self._voltages + \
                    self._currents + \
                    self._z + \
                    self._y + \
                    self._dccurrents + \
                    self._dcvoltages + \
                    self._r + \
                    self._g + \
                    self._times:
                suf = ' #'

            c1 = key + suf
            c2 = self._descr.get(key, '')
            c3 = val
            c4 = self._units.get(key, '-')
            rows.append([c1, c2, c3, c4])

        table.add_rows(rows, header=False)
        table.header(['Parameter', 'Description', 'Default', 'Unit'])

        if export == 'plain':
            pass
        elif export == 'latex':
            raise NotImplementedError('LaTex output not implemented')

        return table.draw()

    def check_limit(self, varname, vmin=None, vmax=None):
        """
        Check if the variable values are within the limits.

        Return False if fails.

        """
        retval = True
        assert varname in self.__dict__

        if varname in self._algebs:
            val = self.system.dae.y[self.__dict__[varname]]
        elif varname in self._states:
            val = self.system.dae.x[self.__dict__[varname]]
        else:  # service or temporary variable
            val = matrix(self.__dict__[varname])

        vmin = matrix(self.__dict__[vmin])
        comp = altb(val, vmin)
        comp = mul(self.u, comp)
        for c, n, idx in zip(comp, self.name, range(self.n)):
            if c == 1:
                v = val[idx]
                vm = vmin[idx]
                self.log(
                    'Init of <{}.{}>={:.4g} is lower than min={:6.4g}'.format(
                        n, varname, v, vm), ERROR)
                retval = False

        vmax = matrix(self.__dict__[vmax])
        comp = agtb(val, vmax)
        comp = mul(self.u, comp)
        for c, n, idx in zip(comp, self.name, range(self.n)):
            if c == 1:
                v = val[idx]
                vm = vmax[idx]
                self.log(
                    'Init of <{}.{}>={:.4g} is higher than max={:.4g}'.format(
                        n, varname, v, vm), ERROR)
                retval = False
        return retval

    def on_bus(self, bus_idx):
        """
        Return the indices of elements on the given buses for shunt-connected
        elements

        :param bus_idx: idx of the buses to which the elements are connected
        :return: idx of elements connected to bus_idx

        """
        assert hasattr(self, 'bus')

        ret = []
        if isinstance(bus_idx, (int, float, str)):
            bus_idx = [bus_idx]

        for item in bus_idx:
            idx = []
            for e, b in enumerate(self.bus):
                if b == item:
                    idx.append(self.idx[e])
            if len(idx) == 1:
                idx = idx[0]
            elif len(idx) == 0:
                idx = None

            ret.append(idx)

        if len(ret) == 1:
            ret = ret[0]

        return ret

    def link_bus(self, bus_idx):
        """
        Return the indices of elements linking the given buses

        :param bus_idx:
        :return:
        """
        ret = []

        if not self._config['is_series']:
            self.log(
                'link_bus function is not valid for non-series model <{}>'.
                format(self.name))
            return []

        if isinstance(bus_idx, (int, float, str)):
            bus_idx = [bus_idx]

        fkey = list(self._ac.keys())
        if 'bus' in fkey:
            fkey.remove('bus')

        nfkey = len(fkey)
        fkey_val = [self.__dict__[i] for i in fkey]

        for item in bus_idx:
            idx = []
            key = []
            for i in range(self.n):
                for j in range(nfkey):
                    if fkey_val[j][i] == item:
                        idx.append(self.idx[i])
                        key.append(fkey[j])
                        # <= 1 terminal should connect to the same bus
                        break

            if len(idx) == 0:
                idx = None
            if len(key) == 0:
                key = None

            ret.append((idx, key))

        return ret

    def elem_find(self, field, value):
        """
        Return the indices of elements whose field first satisfies the given values

        ``value`` should be unique in self.field.
        This function does not check the uniqueness.

        :param field: name of the supplied field
        :param value: value of field of the elemtn to find
        :return: idx of the elements
        :rtype: list, int, float, str
        """
        if isinstance(value, (int, float, str)):
            value = [value]

        f = list(self.__dict__[field])
        uid = np.vectorize(f.index)(value)
        return self.get_idx(uid)

    def _check_Vn(self):
        """Check data consistency of Vn and Vdcn if connected to Bus or Node

        :return None
        """
        if hasattr(self, 'bus') and hasattr(self, 'Vn'):
            bus_Vn = self.read_data_ext('Bus', field='Vn', idx=self.bus)
            for name, bus, Vn, Vn0 in zip(self.name, self.bus, self.Vn,
                                          bus_Vn):
                if Vn != Vn0:
                    self.log(
                        '<{}> has Vn={} different from bus <{}> Vn={}.'.format(
                            name, Vn, bus, Vn0), WARNING)

        if hasattr(self, 'node') and hasattr(self, 'Vdcn'):
            node_Vdcn = self.read_data_ext('Node', field='Vdcn', idx=self.node)
            for name, node, Vdcn, Vdcn0 in zip(self.name, self.node, self.Vdcn,
                                               node_Vdcn):
                if Vdcn != Vdcn0:
                    self.log(
                        '<{}> has Vdcn={} different from node <{}> Vdcn={}.'
                        .format(name, Vdcn, node, Vdcn0), WARNING)

    def link_from(self):
        """

        Returns
        -------

        """
        pass

    def link_to(self, model, idx, self_idx):
        """
        Register (self.name, self.idx) in `model._from`

        Returns
        -------

        """
        if model in self.system.loaded_groups:

            # access group instance
            grp = self.system.__dict__[model]

            # doing it one by one
            for i, self_i in zip(idx, self_idx):
                # query model name and access model instance
                mdl_name = grp._idx_model[i]
                mdl = self.system.__dict__[mdl_name]

                # query the corresponding uid
                u = mdl.get_uid(i)

                # update `mdl_from`
                name_idx_pair = (self._name, self_i)
                if name_idx_pair not in mdl.mdl_from[u]:
                    mdl.mdl_from[u].append(name_idx_pair)

        else:
            # access model instance
            mdl = self.system.__dict__[model]
            uid = mdl.get_uid(idx)

            for u, self_i in zip(uid, self_idx):
                name_idx_pair = (self._name, self_i)

                if name_idx_pair not in mdl.mdl_from[u]:
                    mdl.mdl_from[u].append(name_idx_pair)

    # def var_store_snapshot(self, variable='all'):
    #     """
    #     Store a snapshot of variable values to self._snapshot.
    #
    #     :param variable: name of the variable, or ``all``
    #     :return: None
    #     """
    #
    #     # store (variable name, equation name) in the list ``var_eq_pairs``
    #     # equation name is in ('x', 'y')
    #
    #     var_eq_pairs = []
    #     dae = self.system.dae
    #     if variable != 'all':
    #         if variable in self._states:
    #             var_eq_pairs.append((variable, 'x'))
    #         elif variable in self._algebs:
    #             var_eq_pairs.append((variable, 'y'))
    #         else:
    #             raise NameError('<{}> is not a variable of model {}'
    #                             .format(variable, self._name))
    #     else:
    #         for var in self._states:
    #             var_eq_pairs.append((var, 'x'))
    #         for var in self._algebs:
    #             var_eq_pairs.append((var, 'y'))
    #
    #     # retrieve values and store in ``self._snapshot``
    #
    #     for var, eq in var_eq_pairs:
    #         idx = self.__dict__[var]
    #         self._snapshot[var] = dae.__dict__[eq][idx]



from . import ConfigBase
from ..utils.cached import cached


class Tds(ConfigBase):
    def __init__(self, **kwargs):
        self.fixt = True
        self.tstep = 1 / 30
        self.method = 'trapezoidal'
        self.method_alt = ['euler', 'trapezoidal', 'fwdeuler']
        self.t0 = 0.0
        self.tf = 20
        self.deltat = 0.01
        self.deltatmax = 1
        self.deltatmin = 0.0002
        self.maxit = 30
        self.tol = 1e-4
        self.disturbance = False
        self.error = 1
        self.qrt = False
        self.kqrt = 1
        self.compute_flows = True
        self.max_cache = 0
        super(Tds, self).__init__(**kwargs)

    @cached
    def config_descr(self):
        descriptions = {
            'fixt':
            'use fixed time step size',
            'tstep':
            'time step size',
            'method':
            'time domain integration method',
            't0':
            'starting simulation time',
            'tf':
            'ending simulation time',
            'maxit':
            'maximum iteration number for each integration step',
            'tol':
            'iteration error tolerance',
            'qrt':
            'quasi-real-time simulation speed',
            'kqrt':
            'quasi-rt runs at kqrt seconds per simulated second',
            'compute_flows':
            'post-compute bus injections and line flows at each step',
            'max_cache':
            'maximum allowed steps in varout memory, save to disk if reached, '
            '0 for unlimited memory cache',
        }
        return descriptions



from andes.utils.tab import simpletab
from cvxopt import matrix


def format_newline():
    return '\n'


def format_title(item):
    return item


def format_item(item, val):
    return '{:20s} {:s}'.format(item, str(val))


def format_table(header, data, title=None):
    # fmt = ".4g"
    # return tabulate(data, headers=header, floatfmt=fmt)
    table = simpletab(data=data, header=header)
    return table.draw()
    # return []
    # pass


def dump_data(text, header, rowname, data, file):
    width = 14
    precision = 4
    s = ''
    out = ''
    fid = open(file, 'w')

    for Text, Header, Rowname, Data in zip(text, header, rowname, data):
        # Write Text
        if Text:
            fid.writelines(
                Text
            )

        # Write Header
        if Header:
            ncol = len(Header)
            s = ' ' * width
            s += '{:>{width}s}' * ncol + '\n'
            fid.writelines(s.format(*Header, width=width))  # Mind the asterisk
            fid.write('\n')

        # Append Rowname to Data
        # Data is a list of column lists
        if Rowname:
            ncol = 0
            for idx, item in enumerate(Rowname):  # write by row as always
                if not Data:
                    out = ''
                elif isinstance(Data[0], list):  # list of list in Data
                    ncol = len(Data)
                    out = [Data[i][idx] for i in range(ncol)]
                elif isinstance(Data[0],
                                (int, float)):  # Is just a list of numbers
                    ncol = 1
                    out = [Data[idx]]
                elif isinstance(Data, (int, float)):
                    ncol = 1
                    out = [Data]
                elif isinstance(Data, matrix):  # Data is a matrix
                    pass
                else:
                    print('Unexpected Data during output, in formats/txt.py')

                s = '{:{width}s}'  # for row header
                for col in out:
                    if isinstance(col, (int, float)):
                        s += '{:{width}.{precision}g}'
                    elif type(col) == str:
                        if len(col) > width:
                            col = col[:width]
                        s += '{:{width}s}'
                    else:
                        pass
                s += '\n'

                fid.write(
                    s.format(
                        str(item), *out, width=width, precision=precision))
        fid.write('\n')

    fid.close()



from .base import ModelBase
from cvxopt import matrix, spmatrix, uniform  # NOQA
from ..consts import *  # NOQA
from ..consts import Fx0, Fy0, Gx0, Gy0  # NOQA
from ..consts import Fx, Fy, Gx, Gy  # NOQA


class Bus(ModelBase):
    """ac bus model for defining topology definition
    """

    def define(self):
        self._name = 'Bus'
        self._group = 'Topology'
        self._category = 'Topology'

        self.param_remove('Sn')

        self.param_define(
            'voltage',
            1.0,
            unit='pu',
            descr='initial voltage magnitude',
            nonzero=True)
        self.param_define(
            'angle', 0.0, unit='rad', descr='initial voltage phase angle')
        self.param_define('vmax', 1.1, unit='pu', descr='voltage upper limit')
        self.param_define('vmin', 0.9, unit='pu', descr='voltage upper limit')
        self.param_define('area', 0, descr='area code', tomatrix=False)
        self.param_define('zone', 0, descr='zone code', tomatrix=False)
        self.param_define('region', 0, descr='region code', tomatrix=False)
        self.param_define('owner', 0, descr='owner code', tomatrix=False)
        self.param_define('xcoord', 0, descr='x coordinate', tomatrix=False)
        self.param_define('ycoord', 0, descr='y coordinate', tomatrix=False)

        self.param_alter('Vn', mandatory=True)

        self.var_define('a', 'y', '\\theta', descr='bus voltage phase angle')
        self.var_define('v', 'y', 'V', descr='bus voltage magnitude')

        self.service_define('Pg', matrix)
        self.service_define('Qg', matrix)
        self.service_define('Pl', matrix)
        self.service_define('Ql', matrix)
        self.service_define('islanded_buses', list)
        self.service_define('island_sets', list)

        self._config['address_group_by'] = 'variable'

        self.calls.update({
            'init0': True,
            'pflow': True,
        })

        self._init()

    def _varname_inj(self):
        """Customize varname for bus injections"""
        # Bus Pi
        if not self.n:
            return
        m = self.system.dae.m
        xy_idx = range(m, self.n + m)
        self.system.varname.append(
            listname='unamey',
            xy_idx=xy_idx,
            var_name='P',
            element_name=self.name)
        self.system.varname.append(
            listname='fnamey',
            xy_idx=xy_idx,
            var_name='P',
            element_name=self.name)

        # Bus Qi
        xy_idx = range(m + self.n, m + 2 * self.n)
        self.system.varname.append(
            listname='unamey',
            xy_idx=xy_idx,
            var_name='Q',
            element_name=self.name)
        self.system.varname.append(
            listname='fnamey',
            xy_idx=xy_idx,
            var_name='Q',
            element_name=self.name)

    def init0(self, dae):
        """Set bus Va and Vm initial values"""
        if not self.system.pflow.config.flatstart:
            dae.y[self.a] = self.angle + 1e-10 * uniform(self.n)
            dae.y[self.v] = self.voltage
        else:
            dae.y[self.a] = matrix(0.0,
                                   (self.n, 1), 'd') + 1e-10 * uniform(self.n)
            dae.y[self.v] = matrix(1.0, (self.n, 1), 'd')

    def gisland(self, dae):
        """Reset g(x) for islanded buses and areas"""
        if (not self.islanded_buses) and (not self.island_sets):
            return

        a, v = list(), list()

        # for islanded areas without a slack bus
        # TODO: fix for islanded sets without sw
        # for island in self.island_sets:
        #     nosw = 1
        #     for item in self.system.SW.bus:
        #         if self.uid[item] in island:
        #             nosw = 0
        #             break
        #     if nosw:
        #         self.islanded_buses += island
        #         self.island_sets.remove(island)

        a = self.islanded_buses
        v = [self.n + item for item in a]
        dae.g[a] = 0
        dae.g[v] = 0

    def gyisland(self, dae):
        """Reset gy(x) for islanded buses and areas"""
        if self.system.Bus.islanded_buses:
            a = self.system.Bus.islanded_buses
            v = [self.system.Bus.n + item for item in a]
            dae.set_jac(Gy, 1e-6, a, a)
            dae.set_jac(Gy, 1e-6, v, v)


class BusOld(ModelBase):
    """AC bus model"""

    def __init__(self, system, name):
        """constructor of an AC bus object"""
        super().__init__(system, name)
        self._group = 'Topology'
        self._data.pop('Sn')
        self._data.update({
            'voltage': 1.0,
            'angle': 0.0,
            'vmax': 1.1,
            'vmin': 0.9,
            'area': 0,
            'zone': 0,
            'region': 0,
            'owner': 0,
            'xcoord': None,
            'ycoord': None,
        })
        self._units.update({
            'voltage': 'pu',
            'angle': 'rad',
            'vmax': 'pu',
            'vmin': 'pu',
            'xcoord': 'deg',
            'ycoord': 'deg',
        })
        self._params = [
            'u',
            'Vn',
            'voltage',
            'angle',
            'vmax',
            'vmin',
        ]
        self._descr.update({
            'voltage': 'voltage magnitude in p.u.',
            'angle': 'voltage angle in radian',
            'vmax': 'maximum voltage in p.u.',
            'vmin': 'minimum voltage in p.u.',
            'area': 'area code',
            'zone': 'zone code',
            'region': 'region code',
            'owner': 'owner code',
            'xcoord': 'x coordinate',
            'ycoord': 'y coordinate',
        })
        self._service = ['Pg', 'Qg', 'Pl', 'Ql']
        self._zeros = ['Vn']
        self._mandatory = ['Vn']
        self._algebs.extend(['a', 'v'])
        self._fnamey.extend(['\\theta', 'V'])
        self.calls.update({
            'init0': True,
            'pflow': True,
        })
        self._init()
        self.islanded_buses = list()
        self.island_sets = list()
        self._config['address_group_by'] = 'variable'

    def _varname_inj(self):
        """Customize varname for bus injections"""
        # Bus Pi
        if not self.n:
            return
        m = self.system.dae.m
        xy_idx = range(m, self.n + m)
        self.system.varname.append(
            listname='unamey',
            xy_idx=xy_idx,
            var_name='P',
            element_name=self.name)
        self.system.varname.append(
            listname='fnamey',
            xy_idx=xy_idx,
            var_name='P',
            element_name=self.name)

        # Bus Qi
        xy_idx = range(m + self.n, m + 2 * self.n)
        self.system.varname.append(
            listname='unamey',
            xy_idx=xy_idx,
            var_name='Q',
            element_name=self.name)
        self.system.varname.append(
            listname='fnamey',
            xy_idx=xy_idx,
            var_name='Q',
            element_name=self.name)

    def init0(self, dae):
        """Set bus Va and Vm initial values"""
        if not self.system.pflow.config.flatstart:
            dae.y[self.a] = self.angle + 1e-10 * uniform(self.n)
            dae.y[self.v] = self.voltage
        else:
            dae.y[self.a] = matrix(0.0,
                                   (self.n, 1), 'd') + 1e-10 * uniform(self.n)
            dae.y[self.v] = matrix(1.0, (self.n, 1), 'd')

    def gisland(self, dae):
        """Reset g(x) for islanded buses and areas"""
        if not (self.islanded_buses and self.island_sets):
            return

        a, v = list(), list()

        # for islanded areas without a slack bus
        for island in self.island_sets:
            nosw = 1
            for item in self.system.SW.bus:
                if self.uid[item] in island:
                    nosw = 0
                    break
            if nosw:
                self.islanded_buses += island
                self.island_sets.remove(island)

        a = self.islanded_buses
        v = [self.n + item for item in a]
        dae.g[a] = 0
        dae.g[v] = 0

    def gyisland(self, dae):
        """Reset gy(x) for islanded buses and areas"""
        if self.system.Bus.islanded_buses:
            a = self.system.Bus.islanded_buses
            v = [self.system.Bus.n + item for item in a]
            dae.set_jac(Gy, 1e-6, a, a)
            dae.set_jac(Gy, 1e-6, v, v)



"""
Timed event base class
"""

from .base import ModelBase
from cvxopt import matrix


class EventBase(ModelBase):
    """
    Base class for timed events
    """

    def define(self):
        self.param_remove('Sn')
        self.param_remove('Vn')
        self._group = 'Event'
        self._last_time = -1

    def get_times(self):
        """
        Return a list of occurrance times of the events

        :return: list of times
        """
        if not self.n:
            return list()

        ret = list()

        for item in self._event_times:
            ret += list(self.__dict__[item])

        return ret + list(matrix(ret) - 1e-6)

    def is_time(self, t):
        """
        Return the truth value of whether ``t`` is a time in
        ``self.get_times()``

        :param float t: time to check for
        :return bool: truth value
        """
        if not self.n:
            return False
        else:
            return t in self.get_times()

    def apply(self, sim_time):
        """
        Apply the event and
        :param sim_time:
        :return:
        """
        if not self.n:
            return

        # skip if already applied
        if self._last_time != sim_time:
            self._last_time = sim_time
        else:
            return


class GenTrip(EventBase):
    """
    Timed generator trip and reconnect
    """

    def define(self):
        super(GenTrip, self).define()
        self.param_define(
            'gen',
            default=None,
            mandatory=True,
            descr='generator idx',
            tomatrix=False)
        self.param_define(
            't1',
            default=-1,
            mandatory=False,
            descr='generator trip time',
            event_time=True)
        self.param_define(
            't2',
            default=-1,
            mandatory=False,
            descr='generator reconnect time',
            event_time=True)

        self._init()

    def apply(self, sim_time):
        super(GenTrip, self).apply(sim_time)
        for i in range(self.n):
            gen_idx = self.gen[i]

            if self.t1[i] == sim_time:
                self.log('Applying generator trip on <{}> at t={}'.format(
                    gen_idx, sim_time))
                self.system.Synchronous.set_field('u', gen_idx, 0)
            if self.t2[i] == sim_time:
                self.log('Applying generator reconnect on <{}> at t={}'.format(
                    gen_idx, sim_time))
                self.system.Synchronous.set_field('u', gen_idx, 1)


class LoadShed(EventBase):
    """
    Timed load shedding and reconnect
    """

    def define(self):
        super(LoadShed, self).define()
        self.param_define(
            'load',
            default=None,
            mandatory=True,
            descr='load idx',
            tomatrix=False)
        self.param_define(
            'group',
            default='StaticLoad',
            mandatory=True,
            descr='load group, StaticLoad or DynLoad',
            tomatrix=False)
        self.param_define(
            't1',
            default=-1,
            mandatory=False,
            descr='load shedding time',
            event_time=True)
        self.param_define(
            't2',
            default=-1,
            mandatory=False,
            descr='load reconnect time',
            event_time=True)

        self._init()

    def apply(self, sim_time):
        super(LoadShed, self).apply(sim_time)

        for i in range(self.n):
            load_idx = self.load[i]
            group = self.group[i]

            if self.t1[i] == sim_time:
                self.log('Applying load shedding on <{}> at t={}'.format(
                    load_idx, sim_time))
                self.system.__dict__[group].set_field('u', load_idx, 0)
            if self.t2[i] == sim_time:
                self.log('Applying Gen reconnect on <{}> at t={}'.format(
                    load_idx, sim_time))
                self.system.__dict__[group].set_field('u', load_idx, 1)



from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

from . import main  # NOQA
from . import filters  # NOQA

__author__ = 'Hantao Cui'

__all__ = ['main', 'consts', 'plot', 'system',
           'config', 'filters', 'models', 'routines', 'utils', 'variables']



from cvxopt import umfpack

try:
    from cvxoptklu import klu
    KLU = True
except ImportError:
    KLU = False


class Solver(object):
    """
    Sparse matrix solver class. Wraps UMFPACK and KLU with a unified interface.
    """
    def __init__(self, sparselib='umfpack'):
        self.sparselib = sparselib

    def symbolic(self, A):
        """
        Return the symbolic factorization of sparse matrix ``A``

        Parameters
        ----------
        sparselib
            Library name in ``umfpack`` and ``klu``
        A
            Sparse matrix

        Returns
        symbolic factorization
        -------

        """

        if self.sparselib == 'umfpack':
            return umfpack.symbolic(A)

        elif self.sparselib == 'klu':
            return klu.symbolic(A)

    def numeric(self, A, F):
        """
        Return the numeric factorization of sparse matrix ``A`` using symbolic factorization ``F``

        Parameters
        ----------
        A
            Sparse matrix
        F
            Symbolic factorization

        Returns
        -------
        N
            Numeric factorization of ``A``
        """
        if self.sparselib == 'umfpack':
            return umfpack.numeric(A, F)

        elif self.sparselib == 'klu':
            return klu.numeric(A, F)

    def solve(self, A, F, N, b):
        """
        Solve linear system ``Ax = b`` using numeric factorization ``N`` and symbolic factorization ``F``.
        Store the solution in ``b``.

        Parameters
        ----------
        A
            Sparse matrix
        F
            Symbolic factorization
        N
            Numeric factorization
        b
            RHS of the equation

        Returns
        -------
        None
        """
        if self.sparselib == 'umfpack':
            umfpack.solve(A, N, b)

        elif self.sparselib == 'klu':
            klu.solve(A, F, N, b)

    def linsolve(self, A, b):
        """
        Solve linear equation set ``Ax = b`` and store the solutions in ``b``.

        Parameters
        ----------
        A
            Sparse matrix

        b
            RHS of the equation

        Returns
        -------
        None
        """
        if self.sparselib == 'umfpack':
            return umfpack.linsolve(A, b)

        elif self.sparselib == 'klu':
            return klu.linsolve(A, b)



from cvxopt import matrix, sparse, spmatrix  # NOQA
from cvxopt import mul, div, log, sin, cos  # NOQA

from .base import ModelBase
from ..consts import Fx0, Fy0, Gx0, Gy0  # NOQA
from ..consts import Fx, Fy, Gx, Gy  # NOQA


class GovernorBase(ModelBase):
    """Turbine governor base class"""

    def __init__(self, system, name):
        super(GovernorBase, self).__init__(system, name)
        self._group = 'Governor'
        self.param_remove('Vn')
        self._data.update({
            'gen': None,
            'pmax': 999.0,
            'pmin': 0.0,
            'R': 0.05,
            'wref0': 1.0,
        })
        self._descr.update({
            'gen': 'Generator index',
            'pmax': 'Maximum turbine output in Syn Sn',
            'pmin': 'Minimum turbine output in Syn Sn',
            'R': 'Speed regulation droop',
            'wref0': 'Initial reference speed',
        })
        self._units.update({
            'pmax': 'pu',
            'pmin': 'pu',
            'wref0': 'pu',
            'R': 'pu'
        })
        self._params.extend(['pmax', 'pmin', 'R', 'wref0'])
        self._algebs.extend(['wref', 'pout'])
        self._fnamey.extend(['\\omega_{ref}', 'P_{out}'])
        self._service.extend(['pm0', 'gain'])
        self._mandatory.extend(['gen', 'R'])
        self._powers.extend(['pmax', 'pmin'])
        self.calls.update({
            'init1': True,
            'gcall': True,
            'fcall': True,
            'jac0': True,
        })

    def data_to_sys_base(self):
        if not self.n:
            return
        self.copy_data_ext(
            model='Synchronous', field='Sn', dest='Sn', idx=self.gen)
        super(GovernorBase, self).data_to_sys_base()
        self.R = self.system.mva * div(self.R, self.Sn)

    def init1(self, dae):
        self.gain = div(1.0, self.R)

        # values
        self.copy_data_ext(
            model='Synchronous', field='pm0', dest='pm0', idx=self.gen)

        # indices
        self.copy_data_ext(
            model='Synchronous', field='omega', dest='omega', idx=self.gen)
        self.copy_data_ext(
            model='Synchronous', field='pm', dest='pm', idx=self.gen)

        self.init_limit(
            key='pm0', lower=self.pmin, upper=self.pmax, limit=True)
        dae.y[self.wref] = self.wref0
        dae.y[self.pout] = self.pm0

    def gcall(self, dae):
        dae.g[self.pm] += self.pm0 - mul(
            self.u, dae.y[self.pout])  # update the Syn.pm equations
        dae.g[self.wref] = dae.y[self.wref] - self.wref0

    def jac0(self, dae):
        dae.add_jac(Gy0, -self.u, self.pm, self.pout)

        dae.add_jac(Gy0, 1.0, self.wref, self.wref)


class TG1(GovernorBase):
    """Turbine governor model"""

    def __init__(self, system, name):
        super(TG1, self).__init__(system, name)
        self._name = "TG1"
        self._data.update({
            'T3': 0.0,
            'T4': 12.0,
            'T5': 50.0,
            'Tc': 0.56,
            'Ts': 0.1,
        })
        self._params.extend(['T3', 'T4', 'T5', 'Tc', 'Ts'])
        self._descr.update({
            'T3': 'Transient gain time constant',
            'T4': 'Power fraction time constant',
            'T5': 'Reheat time constant',
            'Tc': 'Servo time constant',
            'Ts': 'Governor time constant',
        })
        self._units.update({
            'T3': 's',
            'T4': 's',
            'T5': 's',
            'Tc': 's',
            'Ts': 's'
        })
        self._mandatory.extend(['T5', 'Tc', 'Ts'])
        self._service.extend(['iTs', 'iTc', 'iT5', 'k1', 'k2', 'k3', 'k4'])
        self._states.extend(['xg1', 'xg2', 'xg3'])
        self._fnamex.extend(['x_{g1}', 'x_{g2}', 'x_{g3}'])
        self._algebs.extend(['pin'])
        self._fnamey.extend(['P_{in}'])
        self._init()

    def init1(self, dae):
        super(TG1, self).init1(dae)
        self.iTs = div(1, self.Ts)
        self.iTc = div(1, self.Tc)
        self.iT5 = div(1, self.T5)
        self.k1 = mul(self.T3, self.iTc)
        self.k2 = 1 - self.k1
        self.k3 = mul(self.T4, self.iT5)
        self.k4 = 1 - self.k3

        dae.x[self.xg1] = mul(self.u, self.pm0)
        dae.x[self.xg2] = mul(self.u, self.k2, self.pm0)
        dae.x[self.xg3] = mul(self.u, self.k4, self.pm0)
        dae.y[self.pin] = self.pm0

    def fcall(self, dae):
        dae.f[self.xg1] = mul(self.u, dae.y[self.pin] - dae.x[self.xg1],
                              self.iTs)
        dae.f[self.xg2] = mul(self.u,
                              mul(self.k2, dae.x[self.xg1]) - dae.x[self.xg2],
                              self.iTc)
        dae.f[self.xg3] = mul(
            self.u,
            mul(self.k4, dae.x[self.xg2] + mul(self.k1, dae.x[self.xg1])) -
            dae.x[self.xg3], self.iT5)

    def gcall(self, dae):
        dae.g[self.pin] = self.pm0 + mul(
            self.gain, dae.y[self.wref] - dae.x[self.omega]) - dae.y[self.pin]
        dae.hard_limit(self.pin, self.pmin, self.pmax)

        dae.g[self.pout] = dae.x[self.xg3] + mul(
            self.k3,
            dae.x[self.xg2] + mul(self.k1, dae.x[self.xg1])) - dae.y[self.pout]
        super(TG1, self).gcall(dae)

    def jac0(self, dae):
        super(TG1, self).jac0(dae)
        dae.add_jac(Gy0, -self.u + 1e-6, self.pin, self.pin)

        dae.add_jac(Gx0, -mul(self.u, self.gain), self.pin, self.omega)
        dae.add_jac(Gy0, mul(self.u, self.gain), self.pin, self.wref)

        dae.add_jac(Fx0, -mul(self.u, self.iTs) + 1e-6, self.xg1, self.xg1)
        dae.add_jac(Fy0, mul(self.u, self.iTs), self.xg1, self.pin)

        dae.add_jac(Fx0, mul(self.u, self.k2, self.iTc), self.xg2, self.xg1)
        dae.add_jac(Fx0, -mul(self.u, self.iTc), self.xg2, self.xg2)

        dae.add_jac(Fx0, mul(self.u, self.k4, self.iT5), self.xg3, self.xg2)
        dae.add_jac(Fx0, mul(self.u, self.k4, self.k1, self.iT5), self.xg3,
                    self.xg1)
        dae.add_jac(Fx0, -mul(self.u, self.iT5), self.xg3, self.xg3)

        dae.add_jac(Gx0, self.u, self.pout, self.xg3)
        dae.add_jac(Gx0, mul(self.u, self.k3), self.pout, self.xg2)
        dae.add_jac(Gx0, mul(self.u, self.k3, self.k1), self.pout, self.xg1)
        dae.add_jac(Gy0, -self.u + 1e-6, self.pout, self.pout)


class TG2(GovernorBase):
    """Simplified governor model"""

    def __init__(self, system, name):
        super(TG2, self).__init__(system, name)
        self._name = 'TG2'
        self._data.update({
            'T1': 0.2,
            'T2': 10.0,
        })
        self._descr.update({
            'T1': 'Transient gain time constant',
            'T2': 'Governor time constant',
        })
        self._units.update({'T1': 's', 'T2': 's'})
        self._params.extend(['T1', 'T2'])
        self._service.extend(['T12', 'iT2'])
        self._mandatory.extend(['T2'])
        self._states.extend(['xg'])
        self._fnamex.extend(['x_g'])
        self._init()

    def init1(self, dae):
        super(TG2, self).init1(dae)
        self.T12 = div(self.T1, self.T2)
        self.iT2 = div(1, self.T2)

    def fcall(self, dae):
        dae.f[self.xg] = mul(
            self.iT2,
            mul(self.gain, 1 - self.T12, self.wref0 - dae.x[self.omega]) -
            dae.x[self.xg])

    def gcall(self, dae):
        pm = dae.x[self.xg] + self.pm0 + mul(self.gain, self.T12,
                                             self.wref0 - dae.x[self.omega])
        dae.g[self.pout] = pm - dae.y[self.pout]
        dae.hard_limit(self.pout, self.pmin, self.pmax)
        super(TG2, self).gcall(dae)

    def jac0(self, dae):
        super(TG2, self).jac0(dae)
        dae.add_jac(Fx0, -self.iT2, self.xg, self.xg)
        dae.add_jac(Fx0, -mul(self.iT2, self.gain, 1 - self.T12), self.xg,
                    self.omega)

        dae.add_jac(Gx0, 1.0, self.pout, self.xg)
        dae.add_jac(Gx0, -mul(self.gain, self.T12), self.pout, self.omega)
        dae.add_jac(Gy0, -1.0, self.pout, self.pout)



from cvxopt import matrix, spmatrix  # NOQA
from cvxopt import mul, div, exp  # NOQA

from .base import ModelBase
from ..consts import Fx0, Fy0, Gx0, Gy0  # NOQA
from ..consts import Fx, Fy, Gx, Gy  # NOQA


class AVR1(ModelBase):
    """
    Automatic Voltage Regulator Type I, DC exciter simplified from IEEE DC1
    """

    def __init__(self, system, name):
        super().__init__(system, name)
        self._group = 'AVR'
        self._name = 'AVR1'
        self._mandatory.extend(['syn'])
        self._states.extend(['vm', 'vr1', 'vr2', 'vfout'])
        self._fnamex.extend(['v_{m}', 'v_{r1}', 'v_{r2}', 'v_{fout}'])
        self._service.extend(['vref0', 'KfTf', 'usyn', 'u0', 'iTa'])
        self._fnamey.extend(['v_{ref}'])
        self._times.extend(['Tr', 'Te', 'Ta', 'Tf'])
        self._algebs.extend(['vref'])
        self._params.extend([
            'Ka', 'Ke', 'Kf', 'Ta', 'Tf', 'Tr', 'Te', 'vrmax', 'vrmin', 'Ae',
            'Be'
        ])
        self._descr.update({
            'busr': 'Regulated voltage idx',
            'vrmax': 'Maximum regulator voltage',
            'Kf': 'Stabilizer gain',
            'Ka': 'Amplifier gain',
            'syn': 'Generator id',
            'Ta': 'Amplifier time constant',
            'vrmin': 'minimum regulator voltage',
            'Tr': 'Measurement time constant',
            'Be': '2nd ceiling coefficient',
            'Tf': 'Stabilizer time constant',
            'Ae': '1st ceiling coefficient',
            'Ke': 'Field circuit integral deviation',
            'Te': 'Field circuit time constant'
        })
        self._data.update({
            'busr': None,
            'vrmax': 5,
            'Kf': 0.063,
            'Ka': 20,
            'syn': 0,
            'Ta': 0.2,
            'vrmin': -5,
            'Tr': 0.001,
            'Be': 0.9,
            'Tf': 0.35,
            'Ae': 0.0006,
            'Ke': 1,
            'Te': 1.0
        })
        self.calls.update({
            'jac0': True,
            'gycall': False,
            'gcall': True,
            'fcall': True,
            'fxcall': True,
            'init1': True
        })
        self._zeros.extend(['Ta', 'Tr'])
        self._init()

    def servcall(self, dae):
        self.copy_data_ext('Synchronous', 'u', dest='usyn', idx=self.syn)
        self.copy_data_ext('Synchronous', 'vf0', dest='vf0', idx=self.syn)
        self.copy_data_ext('Synchronous', 'vf', dest='vf', idx=self.syn)
        self.copy_data_ext('Synchronous', 'v', dest='v', idx=self.syn)
        self.KfTf = mul(self.Kf, div(1, self.Tf))
        self.iTa = div(self.u, self.Ta)
        self.u0 = mul(self.u, self.usyn)

    def init1(self, dae):
        self.servcall(dae)
        dae.x[self.vfout] = dae.y[self.vf]
        self.vref0 = div(mul(dae.y[self.vf], self.Ke + self.Se),
                         self.Ka) + dae.y[self.v]
        dae.y[self.vref] = self.vref0
        dae.x[self.vm] = dae.y[self.v]
        dae.x[self.vr1] = mul(dae.y[self.vf], self.Ke + self.Se)
        dae.x[self.vr2] = -mul(dae.y[self.vf], self.KfTf)
        self.check_limit('vfout', vmin='vrmin', vmax='vrmax')

    def gcall(self, dae):
        dae.g[self.vref] = self.vref0 - dae.y[self.vref]
        dae.g += spmatrix(
            mul(self.u0, self.vf0 - dae.x[self.vfout]), self.vf, [0] * self.n,
            (dae.m, 1), 'd')

    def fcall(self, dae):
        dae.f[self.vm] = mul(div(1, self.Tr), dae.y[self.v] - dae.x[self.vm])
        vr1 = mul(
            self.Ka, dae.y[self.vref] - dae.x[self.vm] - dae.x[self.vr2] - mul(
                self.KfTf, dae.x[self.vfout]))
        dae.f[self.vr1] = div(vr1 - dae.x[self.vr1], self.Ta)
        dae.anti_windup(self.vr1, self.vrmin, self.vrmax)
        dae.f[self.vr2] = mul(
            div(1, self.Tf),
            -dae.x[self.vr2] - mul(self.KfTf, dae.x[self.vfout]))
        dae.f[self.vfout] = mul(
            div(1, self.Te),
            dae.x[self.vr1] - mul(dae.x[self.vfout], self.Ke + self.Se))

    def jac0(self, dae):
        dae.add_jac(Gy0, -1, self.vref, self.vref)
        dae.add_jac(Gx0, -self.u0, self.vf, self.vfout)
        dae.add_jac(Fx0, -div(1, self.Tr), self.vm, self.vm)
        dae.add_jac(Fx0, -mul(self.Ka, self.iTa), self.vr1, self.vm)
        dae.add_jac(Fx0, -mul(self.Ka, self.iTa), self.vr1, self.vr2)
        dae.add_jac(Fx0, -div(1, self.Ta), self.vr1, self.vr1)
        dae.add_jac(Fx0, -mul(self.Ka, self.KfTf, self.iTa), self.vr1,
                    self.vfout)
        dae.add_jac(Fx0, -div(1, self.Tf), self.vr2, self.vr2)
        dae.add_jac(Fx0, -mul(self.KfTf, div(1, self.Tf)), self.vr2,
                    self.vfout)
        dae.add_jac(Fy0, div(1, self.Tr), self.vm, self.v)
        dae.add_jac(Fy0, mul(self.Ka, self.iTa), self.vr1, self.vref)

    def fxcall(self, dae):
        dae.add_jac(Fx, mul(div(1, self.Te), -self.Ke - self.dSe), self.vfout,
                    self.vfout)
        dae.add_jac(Fx, div(self.Ke + self.dSe, self.Te), self.vfout, self.vr1)

    @property
    def Se(self):
        dae = self.system.dae
        vfout = dae.x[self.vfout]
        return mul(self.Ae, exp(mul(self.Be, abs(vfout))))

    @property
    def dSe(self):
        dae = self.system.dae
        vfout = dae.x[self.vfout]
        return mul(self.Ae, exp(mul(self.Be, abs(vfout)))) \
            + mul(self.Ae, self.Be, abs(vfout), exp(mul(self.Be, abs(vfout))))


class AVR2(ModelBase):
    """Automatic Voltage Regulator Type II with fast response and high gain"""

    def __init__(self, system, name):
        super().__init__(system, name)
        self._group = 'AVR'
        self._name = 'AVR2'
        self._states.extend(['vm', 'vr1', 'vr2', 'vfout'])
        self._algebs.extend(['vref', 'vr'])
        self._service.extend(['T21', 'vref0', 'T43'])
        self._fnamey.extend(['v_{ref}', 'v_r'])
        self._fnamex.extend(['v_{m}', 'v_{r1}', 'v_{r2}', 'v_{fout}'])
        self._times.extend(['T1', 'T2', 'T3', 'T4'])
        self._zeros.extend(['Te', 'Tr', 'T1', 'T3'])
        self._mandatory.extend(['syn'])
        self._params.extend([
            'K0', 'T1', 'T2', 'T3', 'T4', 'Tr', 'Te', 'vrmax', 'vrmin', 'Ae',
            'Be'
        ])
        self._data.update({
            'K0': 200,
            'Ae': 0.0006,
            'T3': 0.05,
            'Te': 1.0,
            'syn': 0,
            'vrmin': -5,
            'vrmax': 5,
            'Tr': 0.001,
            'T1': 0.01,
            'T4': 1,
            'Be': 0.9,
            'T2': 0.1
        })
        self._descr.update({
            'K0': 'Regulator gain',
            'vrmin': 'minimum regulator voltage',
            'Ae': '1st ceiling coefficient',
            'Tr': 'Measurement time constant',
            'syn': 'Generator id',
            'T1': 'Regulator zero',
            'Be': '2nd ceiling coefficient',
            'Te': 'Field circuit time constant',
            'vrmax': 'Maximum regulator voltage',
            'T2': 'Regulator Pole'
        })
        self.calls.update({
            'fcall': True,
            'init1': True,
            'gcall': True,
            'fxcall': True,
            'gycall': False,
            'jac0': True
        })
        self._init()

    def servcall(self, dae):
        self.copy_data_ext('Synchronous', 'v', dest='v', idx=self.syn)
        self.copy_data_ext('Synchronous', 'u', dest='usyn', idx=self.syn)
        self.copy_data_ext('Synchronous', 'v', dest='v', idx=self.syn)
        self.copy_data_ext('Synchronous', 'vf', dest='vf', idx=self.syn)
        self.copy_data_ext('Synchronous', 'vf0', dest='vf0', idx=self.syn)
        self.T43 = mul(self.T4, div(1, self.T3))
        self.T21 = mul(self.T2, div(1, self.T1))
        self.u0 = mul(self.u, self.usyn)

    def init1(self, dae):
        self.servcall(dae)
        dae.x[self.vfout] = dae.y[self.vf]
        self.vref0 = dae.y[self.v] + mul(dae.y[self.vf], div(1, self.K0),
                                         1 + self.Se)
        dae.y[self.vref] = self.vref0
        dae.x[self.vm] = dae.y[self.v]
        dae.x[self.vr1] = mul(self.K0, 1 - self.T21,
                              self.vref0 - dae.y[self.v])
        dae.y[self.vr] = mul(dae.y[self.vf], 1 + self.Se)
        dae.x[self.vr2] = mul(
            div(1, self.K0), 1 - self.T43, dae.x[self.vr1] + mul(
                self.K0, self.T21, dae.y[self.vref] - dae.x[self.vm]))

    def gcall(self, dae):
        dae.g[self.vref] = self.vref0 - dae.y[self.vref]
        dae.g[self.vr] = -dae.y[self.vr] + mul(self.K0, dae.x[self.vr2]) + mul(
            self.T43, dae.x[self.vr1] + mul(self.K0, self.T21,
                                            dae.y[self.vref] - dae.x[self.vm]))
        dae.hard_limit(self.vr, self.vrmin, self.vrmax)
        dae.g += spmatrix(
            mul(self.u0, self.vf0 - dae.x[self.vfout]), self.vf, [0] * self.n,
            (dae.m, 1), 'd')

    def fcall(self, dae):
        dae.f[self.vm] = mul(div(1, self.Tr), dae.y[self.v] - dae.x[self.vm])
        dae.f[self.vr1] = mul(
            div(1, self.T1), -dae.x[self.vr1] + mul(
                self.K0, 1 - self.T21, dae.y[self.vref] - dae.x[self.vm]))
        dae.f[self.vr2] = mul(
            div(1, self.K0), div(1, self.T3),
            mul(
                1 - self.T43, dae.x[self.vr1] + mul(
                    self.K0, self.T21, dae.y[self.vref] - dae.x[self.vm])) -
            mul(self.K0, dae.x[self.vr2]))
        dae.f[self.vfout] = mul(
            div(1, self.Te),
            dae.y[self.vr] - mul(dae.x[self.vfout], 1 + self.Se))

    def jac0(self, dae):
        dae.add_jac(Gy0, -1, self.vref, self.vref)
        dae.add_jac(Gy0, mul(self.K0, self.T21, self.T43), self.vr, self.vref)
        dae.add_jac(Gy0, -1, self.vr, self.vr)
        dae.add_jac(Gx0, self.T43, self.vr, self.vr1)
        dae.add_jac(Gx0, self.K0, self.vr, self.vr2)
        dae.add_jac(Gx0, -mul(self.K0, self.T21, self.T43), self.vr, self.vm)
        dae.add_jac(Gx0, -self.u0, self.vf, self.vfout)
        dae.add_jac(Fx0, -div(1, self.Tr), self.vm, self.vm)
        dae.add_jac(Fx0, -div(1, self.T1), self.vr1, self.vr1)
        dae.add_jac(Fx0, -mul(self.K0, div(1, self.T1), 1 - self.T21),
                    self.vr1, self.vm)
        dae.add_jac(Fx0, mul(div(1, self.K0), div(1, self.T3), 1 - self.T43),
                    self.vr2, self.vr1)
        dae.add_jac(Fx0, -div(1, self.T3), self.vr2, self.vr2)
        dae.add_jac(Fx0, -mul(self.T21, div(1, self.T3), 1 - self.T43),
                    self.vr2, self.vm)
        dae.add_jac(Fy0, div(1, self.Tr), self.vm, self.v)
        dae.add_jac(Fy0, mul(self.K0, div(1, self.T1), 1 - self.T21), self.vr1,
                    self.vref)
        dae.add_jac(Fy0, mul(self.T21, div(1, self.T3), 1 - self.T43),
                    self.vr2, self.vref)

    def fxcall(self, dae):
        dae.add_jac(Fy, div(1 + self.dSe, self.Te), self.vfout, self.vr)
        dae.add_jac(Fx, -div(1 + self.dSe, self.Te), self.vfout, self.vfout)

    @property
    def Se(self):
        dae = self.system.dae
        vfout = dae.x[self.vfout]
        return mul(self.Ae, exp(mul(self.Be, abs(vfout))))

    @property
    def dSe(self):
        dae = self.system.dae
        vfout = dae.x[self.vfout]
        return mul(self.Ae, exp(mul(self.Be, abs(vfout)))) \
            + mul(self.Ae, self.Be, abs(vfout), exp(mul(self.Be, abs(vfout))))


class AVR3(ModelBase):
    """Automatic Voltage Regulator Type III"""

    def __init__(self, system, name):
        super().__init__(system, name)
        self._group = 'AVR'
        self._name = 'AVR3'
        self._service.extend(['v0', 'vref0', 'T1T2'])
        self._params.extend(
            ['T1', 'T2', 'Tr', 'Te', 'vfmax', 'vfmin', 'K0', 's0'])
        self._algebs.extend(['vref'])
        self._times.extend(['T2', 'Te', 'Tr'])
        self._fnamex.extend(['v_{m}', 'v_{r}', 'v_{fout}'])
        self._mandatory.extend(['syn'])
        self._states.extend(['vm', 'vr', 'vfout'])
        self._fnamey.extend(['v_{ref}'])
        self._units.update({
            'T1': 's',
            'T2': 's',
            'Te': 's',
            'Tr': 's',
            'vfmax': 'pu',
            'vfmin': 'pu'
        })
        self._data.update({
            'T1': 0.01,
            'T2': 0.1,
            'Tr': 0.001,
            'K0': 20,
            'vfmax': 5,
            'Te': 1.0,
            'syn': 0,
            'vfmin': -5,
            's0': True
        })
        self._descr.update({
            'T1': 'Regulator zero',
            'T2': 'Regularot Pole',
            'Tr': 'Measurement time constant',
            'K0': 'Regulator gain',
            'vfmax': 'Maximum field voltage',
            'Te': 'Field circuit time constant',
            'syn': 'Generator id',
            'vfmin': 'Minimum field voltage',
            's0': 'Enable excitation voltage feedback'
        })
        self.calls.update({
            'fxcall': True,
            'gycall': False,
            'fcall': True,
            'gcall': True,
            'init1': True,
            'jac0': True
        })
        self._init()

    def servcall(self, dae):
        self.copy_data_ext('Synchronous', 'u', dest='usyn', idx=self.syn)
        self.copy_data_ext('Synchronous', 'v', dest='v', idx=self.syn)
        self.copy_data_ext('Synchronous', 'vf', dest='vf', idx=self.syn)
        self.copy_data_ext('Synchronous', 'vf0', dest='vf0', idx=self.syn)
        self.T1T2 = mul(self.T1, div(1, self.T2))
        self.iTe = div(self.u, self.Te)
        self.u0 = mul(self.u, self.usyn)

    def init1(self, dae):
        self.servcall(dae)
        self.v0 = dae.y[self.v]
        self.vref0 = dae.y[self.v]
        dae.y[self.vref] = dae.y[self.v]
        dae.x[self.vfout] = dae.y[self.vf]
        dae.x[self.vm] = dae.y[self.v]

    def gcall(self, dae):
        dae.g[self.vref] = self.vref0 - dae.y[self.vref]
        dae.g += spmatrix(
            mul(self.u0, self.vf0 - dae.x[self.vfout]), self.vf, [0] * self.n,
            (dae.m, 1), 'd')

    def fcall(self, dae):
        dae.f[self.vm] = mul(div(1, self.Tr), dae.y[self.v] - dae.x[self.vm])
        dae.f[self.vr] = mul(
            div(1, self.T2), -dae.x[self.vr] + mul(
                self.K0, 1 - self.T1T2, dae.y[self.vref] - dae.x[self.vm]))
        vfout = mul(
            1 + mul(self.s0, -1 + mul(dae.y[self.v], div(1, self.v0))),
            self.vf0 + dae.x[self.vr] + mul(self.K0, self.T1T2,
                                            dae.y[self.vref] - dae.x[self.vm]))
        dae.f[self.vfout] = div(vfout - dae.x[self.vfout], self.Te)
        dae.anti_windup(self.vfout, self.vfmin, self.vfmax)

    def fxcall(self, dae):
        dae.add_jac(
            Fx,
            -mul(self.iTe, self.K0, self.T1T2,
                 1 + mul(self.s0, -1 + mul(dae.y[self.v], div(1, self.v0)))),
            self.vfout, self.vm)
        dae.add_jac(
            Fx, self.iTe + mul(self.iTe, self.s0,
                               -1 + mul(dae.y[self.v], div(1, self.v0))),
            self.vfout, self.vr)
        dae.add_jac(
            Fy,
            mul(self.iTe, self.K0, self.T1T2,
                1 + mul(self.s0, -1 + mul(dae.y[self.v], div(1, self.v0)))),
            self.vfout, self.vref)
        dae.add_jac(
            Fy,
            mul(
                self.iTe, self.s0, div(1, self.v0),
                self.vf0 + dae.x[self.vr] + mul(
                    self.K0, self.T1T2, dae.y[self.vref] - dae.x[self.vm])),
            self.vfout, self.v)

    def jac0(self, dae):
        dae.add_jac(Gy0, -1, self.vref, self.vref)
        dae.add_jac(Gx0, -self.u0, self.vf, self.vfout)
        dae.add_jac(Fx0, -div(1, self.Tr), self.vm, self.vm)
        dae.add_jac(Fx0, -mul(self.K0, div(1, self.T2), 1 - self.T1T2),
                    self.vr, self.vm)
        dae.add_jac(Fx0, -div(1, self.T2), self.vr, self.vr)
        dae.add_jac(Fy0, div(1, self.Tr), self.vm, self.v)
        dae.add_jac(Fy0, mul(self.K0, div(1, self.T2), 1 - self.T1T2), self.vr,
                    self.vref)
        dae.add_jac(Fx0, -div(1, self.Te), self.vfout, self.vfout)
